/*  CPU.CPP - 65c816 emulation module and main "run" module
	- The main emulation routine is called runsnes ()
	  Allow the user to go to the GUI by pressing ESC.  While in the GUI you can
	  return to the debugger where you left off by clicking Return to Game.
	- The resetsystem () function is supposed to reset the game.
	  This function is called right after a rom is loaded, so you can init the
	  SNES.  It can also be called when you click the Reset button in the Startup
	  dialog box.  This function should not clear the SRAM.
	- When you have time, write the instantload() and instantsave () functions.
	  I already wrote the SRAM save/load function, which automatically decides
	  how many kilobits to save.
	- I made a new array: byte vram [0x10000]; I think the vram is 64K??
	- To set text mode (for your debugger) use:
	  setvideomode (0, false);
	  But if you want, I can make a full-featured GUI debugger.  I haven't written
	  any text functions (eg gotoxy ()).
	- Feel free to use the debug-out function for debugging.  You will notice when you
	  run SNEqr 2 debug files are created called DebugN.doc.  I do all my debug outs to
	  Debug0.doc so if you don't want yours mixed with mine you can use Debug1.doc.
	  The debug-out function is like this:
		void debugf (int file_num, char *format, ...);
	  It's basically just like printf ().  A time stamp is automatically recorded.
*/

#include "Common.hpp"
	// Defines true & false, includes STDIO and other common headers; defines byte,
	// word, dword and boolean data types; and includes MiscFunc.hpp, which contains
	// Timer, keyboard, mouse, joystick and debugging functions.
#include "GrEngine.hpp"
#include "FileMan.hpp"
#include "GUI.hpp"
#include "CPU.hpp"
#include "emu65816.hpp"
#include "SPC700.hpp"
#include "Screen.hpp"
#include <ASSERT.H>

//#define USE_C_CPU_CORE

byte __cdecl *startaddr [0x800];
	/*  For quick address aaccess.  To access a 65c816 address bb.aaaa, stored in variable addr:
		startaddr [addr >> 13] + (addr & 0x1FFF)
		This array is filled by getstartaddresses ().
	*/
byte ram [0x20000], sram [0x10000], expram [0x2000];
byte vram [0x10000];
byte __cdecl registers [0x4000];
byte cgram [512], oam [544];
byte *rom = NULL;

boolean debugmode = false;
boolean returntogui;
boolean showfps = true;

int snesframe = 0;
int snesrenderedframe = 0;

int instantsaveslot = 0;

int frameskipwait = 0;

int autoback_countdown = 0, autoback_restorecount;

struct headerdata curheader;
	// This is the header data that has been extracted from the ROM image that is currently
	// loaded.

struct ppuinternalregisters state;

byte *autobacktrackmem[10] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
int autobackpos[9], autobacksize[9];

void bcdaddadjust_byte (byte &A)
{
	if ((A & 15) > 9) A += 6;
	if (((A >> 4) & 15) > 9) A += (6 << 4);
}
void bcdsubadjust_byte (byte &A)
{
	if ((A & 15) > 9) A -= 6;
	if (((A >> 4) & 15) > 9) A -= (6 << 4);
}
void bcdaddadjust_word (word &A)
{
	if ((A & 15) > 9) A += 6;
	if (((A >> 4) & 15) > 9) A += (6 << 4);
	if (((A >> 8) & 15) > 9) A += (6 << 8);
	if (((A >> 12) & 15) > 9) A += (6 << 12);
}
void bcdsubadjust_word (word &A)
{
	if ((A & 15) > 9) A -= 6;
	if (((A >> 4) & 15) > 9) A -= (6 << 4);
	if (((A >> 8) & 15) > 9) A -= (6 << 8);
	if (((A >> 12) & 15) > 9) A -= (6 << 12);
}

int debug_slice = -1;
long __cdecl debug_instr = 0;
int __cdecl frame_cycles, scan_cycles, screen_scanline, total_frames, cycles_per_scan;

void emulate_m0_x0 ();
void emulate_m0_x1 ();
void emulate_m1_x0 ();
void emulate_m1_x1 ();
// Create different versions of the emulation routine.
// These functions return the amount of cycles left.  If number is negative, the number
// of cycles "over" that occured.

int debug_trapaddr = -1; boolean debug_readtrap;

byte debug_thing;
int instr_tally [256][2];
int mmio_rtally [768];
int mmio_wtally [768];

void savetallies ()
{
#ifndef NODEBUGOUT
	char *fn, ext[4] = "DOC";
	FILE *fp;
	int x, y, c;
	char *s;
	boolean usedlist [256];

	fn = getaccessoryname (curgs->filename, instdir, ext);
	if ((fp = fopen (fn, "wt")) == NULL) {
		addmessage ("Couldn't open tally file %s...", fn);
		return;
	}
	fprintf (fp, "Addressing Mode Counts:\n");
	memset (usedlist,false,sizeof(usedlist));
	for (x = 0; x < 256; x++) {
		if (usedlist[x])
			continue;
		for (y = x, c = 0; y < 256; y++) {
			if (stricmp (opcodelist[x].format+3, opcodelist[y].format+3) == 0) {
				usedlist[y] = true;
				c += instr_tally[y][0] + instr_tally[y][0];
			}
		}
		fprintf (fp, "%d\t%s\n", c, opcodelist[x].format+3);
	}
	fprintf (fp, "\n\nOpcode Counts:\n");
	for (x = 0; x < 256; x++) {
		s = opcodelist[x].format;
		if ((s[0]=='B' && s[1]=='N' && s[2]=='E') || \
			(s[0]=='B' && s[1]=='E' && s[2]=='Q') || \
			(s[0]=='S' && s[1]=='T') || \
			(s[0]=='L' && s[1]=='D') || \
			(s[0]=='I' && s[1]=='N'))
				continue;
		if ((s[0]=='A' && s[1]=='D' && s[2]=='C') ||\
			(s[0]=='S' && s[1]=='B' && s[2]=='C')) {
			fprintf (fp, "%d/%d\t%s\n", instr_tally[x][0],instr_tally[x][1], s);
		} else {
			fprintf (fp, "%d\t%s\n", instr_tally[x][0]+instr_tally[x][1], s);
		}
	}
	fprintf (fp, "\n\nMMIO Read Tallies\n");
	for (x = 0; x < 256; x++) {
		if (mmio_rtally[x])
			fprintf (fp, "0x%X:%d\t", 0x2100 + x, mmio_rtally[x]);
	}
	fprintf (fp, "\n");
	for (x = 0; x < 512; x++) {
		if (mmio_rtally[x + 256])
			fprintf (fp, "0x%X:%d\t", 0x4200 + x, mmio_rtally[x + 256]);
	}
	fprintf (fp, "\n\nMMIO Write Tallies\n");
	for (x = 0; x < 256; x++) {
		if (mmio_wtally[x])
			fprintf (fp, "0x%X:%d\t", 0x2100 + x, mmio_wtally[x]);
	}
	fprintf (fp, "\n");
	for (x = 0; x < 512; x++) {
		if (mmio_wtally[x + 256])
			fprintf (fp, "0x%X:%d\t", 0x4200 + x, mmio_wtally[x + 256]);
	}
	fclose (fp);

#endif
}

#define CARRY 1
#define ZERO 2
#define INT_DISABLE 4
#define BCD 8
#define INDEX 16
#define MEMORY 32
#define OVERFLOW 64
#define NEGATIVE 128

#ifdef USE_C_CPU_CORE
	#define A_16BIT
	#define XY_16BIT
	#define EMUROUTINE emulate_m0_x0
	#include "65816emu.cpp"
	#undef A_16BIT
	#define XY_16BIT
	#undef EMUROUTINE
	#define EMUROUTINE emulate_m1_x0
	#include "65816emu.cpp"
	#undef XY_16BIT
	#undef EMUROUTINE
	#define EMUROUTINE emulate_m1_x1
	#include "65816emu.cpp"
	#define A_16BIT
	#undef EMUROUTINE
	#define EMUROUTINE emulate_m0_x1
	#include "65816emu.cpp"
#endif

void setjoypadreg (int joypad)
{
//      static int toggle = 0;
	struct snesjoypad *joy = &joykey [pcontrol[joypad]];

	*REG4218(joypad) = (keydown[joy->b] << 15) | (keydown[joy->y] << 14) | (keydown[joy->select] << 13) \
		| (keydown[joy->start] << 12) | (keydown[joy->up] << 11) | (keydown[joy->down] << 10) \
		| (keydown[joy->left] <<  9) | (keydown[joy->right] <<  8) | (keydown[joy->a] << 7) \
		| (keydown[joy->x] <<  6) | (keydown[joy->l] <<  5) | (keydown[joy->r] << 4);
//	toggle++;
//	*REG4218(joypad) = ((toggle & 2) << 6) | ((screen_scanline & 2) << 14); // a & b
		// debug thing
}

boolean getoldjoypadbit (boolean p2, int bit)
{
	int player = p2 + (bit >= 8 ? 2 : 0);
	struct snesjoypad *joy = &joykey [pcontrol[player]];

	if (*REG4200 & 1) {
		if (pcontrol[p2] != 3) return 1;
		else return 0;
	} else if (bit >= 32) {
		switch (bit) {
		case 34: return (pcontrol[p2]   == 3 ? 0 : 1);
		case 35: return (pcontrol[p2+1] == 3 ? 0 : 1);
		default: return 1;
		}
	} else {
		switch (bit & 0xF) {
		case  0: return keydown[joy->b];
		case  1: return keydown[joy->y];
		case  2: return keydown[joy->select];
		case  3: return keydown[joy->start];
		case  4: return keydown[joy->up];
		case  5: return keydown[joy->down];
		case  6: return keydown[joy->left];
		case  7: return keydown[joy->right];
		case  8: return keydown[joy->a];
		case  9: return keydown[joy->x];
		case 10: return keydown[joy->l];
		case 11: return keydown[joy->r];
		default: return 0;
		}
	}
}

byte _cdecl near *trapregread (byte *ptr, boolean wordread)
{
	int x, n;

	if (ptr-registers+0x2000 >= 0x4200 && ptr-registers+0x2000 <= 0x43FF)
		mmio_rtally [ptr-registers+0x2000-0x4200+256]++;
	if (ptr-registers+0x2000 >= 0x2100 && ptr-registers+0x2000 <= 0x21FF)
		mmio_rtally [ptr-registers+0x2000-0x2100]++;
	debug_readtrap = true; debug_trapaddr = ptr - registers + 0x2000;

	switch (ptr - registers + 0x2000) {
	case 0x2138: // OAM read & inc
		*ptr = oam [state.oampointer];
		state.oampointer = (state.oampointer + 1) % 544;
		if (state.oampointer == 0) debug0 ("SNES Program caused wrap of oampointer to 0 (on read)");
		if (state.oampointer == 1) debug0 ("oampointer is now 1 (on read)");
		break;
	case 0x2139: // VRAM read low
		(*(word*)REG2139) = *(word*)(&vram [state.vrampointer]);
		if (state.dummyread2139) {
			state.dummyread2139 = false;
		} if ((*REG2115 & 0x80) == 0) {
			state.vrampointer = (state.vrampointer + state.vramincrate) & 0xFFFF;
		}
		break;
	case 0x213A: // VRAM read high
		(*(word*)REG2139) = *(word*)(&vram [state.vrampointer]);
		//if (*REG2115 & 0x80) {
		//	state.vrampointer = (state.vrampointer + state.vramincrate) & 0xFFFF;
		//}
		break;
	case 0x213B: // CGRAM (palette) read
		*ptr = cgram [state.colorpointer];
		state.colorpointer = (state.colorpointer + 1) % 512;
		if (state.colorpointer == 0) debug0 ("SNES Program caused wrap of colorpointer to 0 (on read)");
		if (state.colorpointer == 1) debug0 ("colorpointer is now 1 (on read)");
		break;
	case 0x213E: // PPU status flag/version number b7:time over b6:range over b5:master/slave b3-0:version
		*ptr = 1;
		break;
	case 0x213F: // PPU status flag/version number b7:field# in int.mode b6:ext.signal b4:1=PAL b0-3:version
		// SNES type
		*ptr = (ispal() << 4) + 1;
		break;
	case 0x2140: case 0x2141: case 0x2142: case 0x2143:
		{ static boolean recursed = false, skipreg;

		if (curgs->emulatespc) {
		} else {
			if (recursed) { // Prevent SPC skip from being applied to high byte
				// of a word read.
				recursed = false;
				break;
			} else if (wordread) {
				recursed = true;
			}
			switch (skipreg) {
			case 0:
				if (reg.P & 0x20) *ptr = reg.A_lo;
				else *(word*)ptr = reg.A;
				break;
			case 1:
				if (reg.P & 0x10) *ptr = reg.X_lo;
				else *(word*)ptr = reg.X;
				break;
			case 2:
				if (reg.P & 0x10) *ptr = reg.Y_lo;
				else *(word*)ptr = reg.Y;
				break;
			}
			skipreg = (skipreg + 1) % 3;
			if (curgs->spcskipmethod == 0) {
				#ifdef USE_C_CPU_CORE
				x = reg.PC + opcodelist[*SNESMEM(reg.PC)].bytes;
				#else
				x = reg.PC;
				#endif
				n = *SNESMEM(x);
				#define ISCONDITIONALOP(n) ((n) == 0x10 || (n) == 0x30 || (n) == 0xD0 || (n) == 0x90 || (n) == 0xB0 || (n) == 0xF0)
				if (ISCONDITIONALOP(n) && (*SNESMEM(x+1) & 0x80)) {
					// Cond. Branch backward
					(*(word*)SNESMEM(x)) = 0xEAEA; // two NOPs
				} else {
					x += opcodelist[n].bytes;
					n = *SNESMEM(x);
					if (ISCONDITIONALOP(n) && (*SNESMEM(x+1) & 0x80)) {
						(*(word*)SNESMEM(x)) = 0xEAEA;
					} else if ((reg.P & 0x20) == 0) {
						x++;
						n = *SNESMEM(x);
						if (ISCONDITIONALOP(n) && (*SNESMEM(x+1) & 0x80))
							(*(word*)SNESMEM(x)) = 0xEAEA;
					}
				}

				if (ptr != ((byte*)REG2140) + 3)
					(*(word*)ptr) = reg.A;
				else
					*ptr = (byte) reg.A;
			} else {
				if (screen_scanline & 2) { // Alternate method, used half the time...
					// Possibly useful if method before ifs doesn't work
					if (screen_scanline & 1) {
						*SNESMEM(0x2140) = (byte)rand();
						*SNESMEM(0x2141) = (byte)rand();
						*SNESMEM(0x2142) = (byte)rand();
						*SNESMEM(0x2143) = (byte)rand();
					} else {
						*(word*)SNESMEM(ptr - registers + 0x2000) = reg.A;
					}
				}
			}
		}

		}
		break;
	case 0x2180: // WRAM data read
		// Accesses only RAM?
		x = ((*REG2181) & 0x1FFFF);
		*REG2180 = ram[x];
		(*(word*)REG2181)++;

		//x = (*REG2181) & 0xFFFFFF;
		//*REG2180 = *SNESMEM(x);
		//(*(word*)REG2181)++;
		if (*(word*)REG2181 == 0) // inc bank or not?
			(*(((byte*)REG2181) + 2))++;
		break;
	case 0x2137: // H/V counter latch
		state.hcounter = scan_cycles;
		state.vcounter = screen_scanline;
		break;
	case 0x213C: // Horizontal counter
		*REG213C = (byte)state.hcounter;
		x = (byte)state.hcounter;
		state.hcounter = (state.hcounter >> 8) + (x << 8);
		break;
	case 0x213D: // Vertical counter
		*REG213D = (byte)state.vcounter;
		x = (byte)state.vcounter;
		state.vcounter = (state.vcounter >> 8) + (x << 8);
		break;
	case 0x4210: // Clear NMI int request.  zSNES often sets the bit only once--at scan 224-regardless of whether NMIs are enabled.  Reading once clears the register.
		//if (screen_scanline < scans_before_vbl()) //|| (*REG4200 & 0x80) == 0)
		//	reg4210 = 1;
		//else
		//	reg4210 = 0x81;
		*REG4210 = state.reg4210;
		state.reg4210 &= ~0x80;
		break;
	case 0x4211:
		*REG4211 = state.reg4211;
		state.reg4211 &= ~0x80;
		break;
	case 0x4212: // V/HBlank/Joypad state
		*REG4212 = (scan_cycles < 40 ? 0x40 : 0) | (screen_scanline >= scans_before_vbl() ? 0x80 : 0) | (screen_scanline & 1);
		if (debugmode && (tmsec & 1)) // To break infinite hblank check loops in debugger
			*REG4212 &= ~0x40;
		break;
	/*case 0x4218: case 0x4219:
	case 0x421A: case 0x421B:
	case 0x421C: case 0x421D:
	case 0x421E: case 0x421F:
		setjoypadreg ((ptr - registers - 0x2218) / 2);
		break;*/
	case 0x4016:
		*REG4016 = getoldjoypadbit (false, state.oldjoybit[false]);
		state.oldjoybit[false]++;
		break;
	case 0x4017:
		*REG4017 = getoldjoypadbit (true, state.oldjoybit[true]);
		state.oldjoybit[true]++;
		break;
	}
	if (wordread)
		trapregread (ptr + 1, false);
	return ptr;
}

inline void changevram (int where, byte what)
{
	if (vram [where] == what)
		return;
	vram [where] = what;
	cstatus [where >> 4] = 0;
	// The following checks are usually redundant, but not always...
	if ((where & 0x10) && (cstatus [(where >> 4) - 1] & 3) == 1)
		cstatus [(where >> 4) - 1] = 0;
	if ((where & 0x30) == 0x20 && (cstatus [(where >> 4) - 2] & 3) == 3)
		cstatus [(where >> 4) - 2] = 0;
	else if ((where & 0x30) == 0x30 && (cstatus [(where >> 4) - 3] & 3) == 3)
		cstatus [(where >> 4) - 3] = 0;
}

byte _cdecl near *trapregwrite (byte *ptr, boolean wordwrite)
{
	int x;

	if (ptr-registers+0x2000 >= 0x4200 && ptr-registers+0x2000 <= 0x43FF)
		mmio_wtally [ptr-registers+0x2000-0x4200+256]++;
	if (ptr-registers+0x2000 >= 0x2100 && ptr-registers+0x2000 <= 0x21FF)
		mmio_wtally [ptr-registers+0x2000-0x2100]++;
	debug_readtrap = false; debug_trapaddr = ptr - registers + 0x2000;

	switch (ptr - registers + 0x2000) {
	case 0x2100:
		break;
	case 0x2101:
		if (*REG2101 != state.reg2101) {
			state.reg2101 = *REG2101;
			memset (spr8x8size, 0, sizeof (spr8x8size));
		}
		break;
	case 0x2102: case 0x2103: // OAM address
		state.oampointer = *REG2102;
		break;
	case 0x2104: // OAM Data Write
		if (oam[state.oampointer] != *ptr) {
			if (state.oampointer >= 512) {
				spr8x8size [((state.oampointer - 512) << 2)] =
				spr8x8size [((state.oampointer - 512) << 2) + 1] =
				spr8x8size [((state.oampointer - 512) << 2) + 2] =
				spr8x8size [((state.oampointer - 512) << 2) + 3] = 0;
				oamposchange = true;
			} else if (state.oampointer & 2) {
				spr8x8size [state.oampointer >> 2] = 0;
			} else {
				oamposchange = true;
			}
		}
		oam[state.oampointer] = *ptr;
		state.oampointer = (state.oampointer + 1) % 544;
		break;
	case 0x2105: // Screen mode register
		break;
	case 0x210D: case 0x210F: case 0x2111: case 0x2113:
	case 0x210E: case 0x2110: case 0x2112: case 0x2114:
		x = ptr - registers - 0x010D;
		state.scrollreg[x] = (state.scrollreg[x] >> 8) | (*ptr << 8);
		break;
	case 0x2115: // VRAM upload control register; b7:0=inc at 2118?
		switch (*REG2115 & 0x03) {
		case 0: state.vramincrate = 2; break;
		case 1: state.vramincrate = 64; break;
		case 2: state.vramincrate = 128; break;
		case 3: state.vramincrate = 256; break;
		}
		break;
	case 0x2116: case 0x2117: // VRAM address register
		state.vrampointer = *REG2116 * 2;
		state.vrampointer &= 0xFFFF;
		state.dummyread2139 = true;
		break;
	case 0x2118: // VRAM Write register low
		changevram (state.vrampointer, *(byte*)REG2118);
		if (!(*REG2115 & 0x80)) {
			state.vrampointer = ((state.vrampointer + state.vramincrate) & 0xFFFF);
			*REG2118 = 0;
		}
		break;
	case 0x2119: // VRAM Write register high
		changevram (state.vrampointer + 1, *((byte*)REG2118 + 1));
		if (*REG2115 & 0x80) {
			state.vrampointer = ((state.vrampointer + state.vramincrate) & 0xFFFF);
			*REG2118 = 0;
		}
		break;
	case 0x211B: case 0x211C:
	case 0x211D: case 0x211E:
	case 0x211F: case 0x2120:
		if (ptr == REG211B) { // Mode 7 multiply, 16-bit value
			state.reg211B = (((unsigned short)state.reg211B) >> 8) + (*ptr << 8);
		} else if (ptr == REG211C) {
			state.reg211C = *REG211C;
			x = state.reg211B * state.reg211C;
			*REG2134 = x;
		}
		state.m7var[ptr-REG211B] = (((unsigned short)state.m7var[ptr-REG211B]) >> 8) | (*ptr << 8);
		break;
	case 0x2121: // Palette (color #) selection register
		state.colorpointer = *REG2121 * 2;
		break;
	case 0x2122: // Palette data write register
		cgram [state.colorpointer] = *REG2122;
		state.colorpointer = (state.colorpointer + 1) & 0x1FF;
		break;
	case 0x212C: // Main screen designation
		break;
	case 0x212D: // Sub screen designation
		break;
	case 0x2131: // Addition/Subtraction for screens, BGs and OBJs
		for (x = 0; x < 6; x++) {
			if ((*REG2131 >> x) & 1)
				state.reg2131setting[x] = (*REG2131 & 192);
		}
		break;
	case 0x2133: // Screen mode
		break;
	case 0x2180: // WRAM data
		// Accesses only RAM?

		x = (*REG2181) & 0x1FFFF;
		ram[x] = *REG2180;
		(*(word*)REG2181)++;
		if (*(word*)REG2181 == 0)
			(*(((byte*)REG2181) + 2))++;

		//x = (*REG2181) & 0xFFFFFF;
		//*SNESMEM(x) = *REG2180;
		//(*(word*)REG2181)++;

		break;
	case 0x4016:
	case 0x4017:
		// New theory: * Writing to either 16 or 17 resets BOTH joypads;
		// * It is only reset properly if non-zero is written followed by 0.
		// * If *only* a 0 is written, the next returned value on read will be 1.
		// * This sets oldjoybit to -1, so return 1 if oldjoybit < 0.
		// ...wait, nope, forget that theory.
		for (x = 0; x < 2; x++) {
			state.oldjoybit[x] = 0;
		/*	if (*ptr == 0 && state.oldjoybit[x] == -2)
				state.oldjoybit[x] = 0;
			else if (*ptr == 0 && state.oldjoybit[x] != -2)
				state.oldjoybit[x] = -1;
			else if (*ptr != 0)
				state.oldjoybit[x] = -2;*/
		}
		break;
	case 0x4200: // Counter enable
		break;
	case 0x4201: // I/O Port out
		*REG4213 = *REG4201;
		break;
	case 0x4206: // It's important (Metroid 3) NOT to trap 4204/5.
		if (*REG4206 != 0) { // Not divide by zero
			*REG4214 = (*REG4204 / *REG4206); // Unsigned division
			*REG4216 = (*REG4204 % *REG4206);
		} else {
			debug0 ("Divide by 0 - %X", ptr - registers + 0x2000);
		}
		break;
	case 0x4203: // Multiplicand 'B' - written second
		*REG4216 = (word)*REG4202 * (word)*REG4203;
		break;
	case 0x420B:
		{ word dmacount, modeoffset;
		byte regoffset;
		dword cpu, reg;
		char debugline [50];

		//debug0 ("DMA transfer initiated; PC=$%X instr=%d", ::reg.PC, debug_instr);
		for (x = 0; x < 8; x++) {
			if (!(*REG420B & ~(*REG420C) & (1 << x)))
				continue;
			dmacount = *REG4305(x);
			modeoffset = regoffset = 0;
			cpu = *REG4302(x) & 0xFFFFFF;
			reg = 0x2100 | *REG4301(x);
			//debug0 ("DMA transfer on channel %d cpu=$%X reg=$%X dmacount=%d $43?0=$%X state.vramincrate%d", x, cpu, reg, dmacount, *REG4300(x), state.vramincrate);
			if (debugmode) {
				sprintf (debugline, "[DMA ch%d mode:%X a$%X b$%X cnt$%X]", x, *REG4300(x), cpu, reg, dmacount);
				//sprintf (debugline, "[reg$%X]", reg);
				//debug0 ("reg:%X", reg);
				adddebugline (debugline, false, '*');
			}
			do {
				switch (*REG4300(x) & 7) {
				case 1: regoffset = modeoffset & 1; break;
				case 3: regoffset = (modeoffset >> 1) & 1; break;
				case 4: regoffset = modeoffset & 3; break;
				}
				if (*REG4300(x) & 0x80) {
					*SNESMEM(cpu) = *trapregread (SNESMEM(reg+regoffset), false);
				} else {
					*SNESMEM(reg+regoffset) = *SNESMEM(cpu);
					trapregwrite (SNESMEM(reg+regoffset), false);
				}
				if (!(*REG4300(x) & 8)) {
					if (*REG4300(x) & 0x10) // decrement
						cpu = (cpu & 0xFF0000) | ((cpu - 1) & 0xFFFF);
					else // Increment
						cpu = (cpu & 0xFF0000) | ((cpu + 1) & 0xFFFF);
				}
				dmacount--;
				modeoffset++;
				cpu &= 0xFFFFFF;
			} while (dmacount != 0);
			*REG4302(x) = (*REG4302(x) & 0xFF000000) | cpu;
			*REG4305(x) = dmacount; //0
		}
		*REG420B = 0;
		}
		break;
	}
	if (wordwrite)
		trapregwrite (ptr + 1, false);

	return ptr;
}

/*
Known potential CPU problems
Wraparound problems:
	When a reference is made that crosses a bank boundary, often the wraparound does NOT
	occur.  Some wraparounds are implemented (e.g. stack decrement), but most are not
	(e.g. pushing a word onto the stack with one byte at 0 and the other at 0xFFFF, will
	screw up, and so will doing a relative long BRL accross a bank boundary).
Memory-mapped I/O problems:
	All memory accesses are supposed to be trapped for full emulation, but for simplicity
	and speed, many accesses are NOT trapped.  Here's a list of known non-traps:
	PC - opcode+data fetches are not intercepted
	S - Stack operations (PUSH and PULL) are not intercepted (although stack relative
		operations are.)
	Indirection - The final address resulting from an indirection is intercepted, but
		the pointer variable itself is not.  e.g.: LDA ($2139)
		This cannot trigger the VRAM-read data mechanism.  But, if this is used: LDA
		(<$0) and $0 contains: $39 $21 then the VRAM read/increment will occur.
	RMW operations: Currently, when a read-modify-write operation occurs (e.g. ROL $xxxx),
		only either the read is trapped or the write is trapped.  Certain operations (like
		INC,DEC) only check the read and some operations (ROL,SHR) only check the write.
		I know, life sucks.  Also, when a write is checked, the flags are set *after* the
		write intercept.  I don't know if this is the correct behavior.
	Block move instructions: block move instructions are not intercepted.
	Indirect jump: Indirect jumps are not intercepted.
ROM:
	- ROM Writes are NOT prevented
	- FastROM handling is a bit iffy
SRAM: The header information is NOT checked for SRAM size, and reads/writes are NOT
	intercepted - a game can always read and write to the full 512 kbits of SRAM address
	space.
Invalid addresses: I can't IMAGINE what will happen if you try to access an invalid
	address (GPF probably...)
Debugger problems: You should only attempt to view one memory area at a time.  For example,
	you shouldn't try to look at the place where the register address space meets up with
	the scratchpad memory space (e.g. M$1FF0)
*/

struct instructioninfo {
	char mnemonic[4];
	char *descr;
} instrlist [] = {
	"ADC", "Add memory to accumulator with carry",
	"AND", "AND arg with accumulator (A&=arg)",
	"SHL", "(ASL) Shift left arg (C set to MSB)",
	"BCC", "Branch if Carry Clear",
	"BCS", "Branch if carry set",
	"BEQ", "Branch if zero (or branch if equal)",
	"BIT", "And A w/ arg, N=bit 7, V=bit 6, discard result",
	"BMI", "Branch if minus (if N=1)",
	"BNE", "Branch if not zero (branch if not equal)",
	"BPL", "Branch if plus (if N=0)",
	"BRA", "Branch always (Relative)",
	"BRK", "Break (Push PC+2; Call Interrupt *0.FFE6)",
	"BRL", "Branch Long (16-bit JMP)",
	"BVC", "Branch if Overflow Clear",
	"BVS", "Branch if overflow set",
	"CLC", "Clear the Cary Bit (P&~1)",
	"CLD", "Clear the BCD mode bit (P&~8)",
	"CLI", "Clear the Interrupt disable bit (P&~4)",
	"CLV", "Clear the overflow bit (P&~64)",
	"CMP", "Compare arg with A: A>=a?C=1:N=1,A=a?Z=1:Z=0",
	"CSP", "(COP) Call Sys.Proc. (Push PC; Call Int. *0.FFE4)",
	"CPX", "Compare arg with X",
	"CPY", "Compare arg with Y",
	"DEC", "Decrement arg (A or a memory location)",
	"DEX", "Decrement X",
	"DEY", "Decrement Y",
	"XOR", "(EOR) Exclusive OR A with arg (A^=arg)",
	"INC", "Increment arg (A or a memory location)",
	"INX", "Increment X",
	"INY", "Increment Y",
	"JML", "Jump Long (PC=16bit arg->24bit address)",
	"JMP", "Jump (PC=arg or PC=*arg)",
	"JSL", "Jump to Subroutine Long (PC=24bit arg)",
	"JSR", "Jump to Subroutine",
	"LDA", "Load A with a value",
	"LDX", "Load Index X with a value",
	"LDY", "Load Index Y with a value",
	"SHR", "(LSR) Shift Right (C set to LSB)",
	"MVN", "Move- (DBR=arg;while(A){*Y=*arg2.X;X++,Y++,A--})",
	"MVN", "Move+ (DBR=arg;while(A){*Y=*arg2.X;X--,Y--,A--})",
	"NOP", "No operation--waste two cycles",
	"ORA", "OR arg with accumulator (A|=arg)",
	"PSH", "(PE?,PH?) Push onto the stack",
	"PUL", "(PL?) Pull from the stack",
	"CLP", "(REP) Clear status bits in P",
	"ROL", "Rotate arg left (carry included)",
	"ROR", "Rotate arg right (carry included)",
	"RTI", "Return from interrupt & restore flags",
	"RTL", "Return from Subroutine Long",
	"RTS", "Return from Subroutine",
	"SBC", "Subtract from Accumulator with Carry (Borrow)",
	"SEC", "Set the Carry bit to 1",
	"SED", "Set the BCD mode bit to 1",
	"SEI", "Set the Interrupt Disable bit to 1",
	"SEP", "Set status bits in P",
	"HLT", "(STP-Stop the clock) Halt CPU until reset",
	"STA", "Store Accumulator in memory",
	"STX", "Store Index X in memory",
	"STY", "Store Index Y in memory",
	"CLR", "(STZ) Clear memory location",
	"TRB", "Test and reset bits",
	"TSB", "Test and set bits",
	"TAX", "Transfer (copy) A into X",
	"TXA", "Transfer (copy) X into A",
	"TAY", "Transfer (copy) A into Y",
	"TYA", "Transfer (copy) Y into A",
	"TAD", "Transfer (copy) A into Direct",
	"TDA", "Transfer (copy) Direct into A",
	"TAS", "Transfer (copy) A into Stack Pointer",
	"TSA", "Transfer (copy) Stack pointer into A",
	"TSX", "Transfer (copy) Stack pointer into X",
	"TXS", "Transfer (copy) X into Stack pointer",
	"TXY", "Transfer (copy) X into Y",
	"TYX", "Transfer (copy) Y into X",
	"WAI", "Wait for interrupt",
	"SWA", "(XBA) Swap 16-bit accumulator halves",
	"XCE", "Exchange the Carry and Emulation bits",
	"???", "UNASSIGNED OPCODE $42",
	"\0\0\0", "[Description search error]"
};

struct opcodeinfo opcodelist [] = {
	// 0x0
	"BRK", 2, 8,
	"ORA (<%b,X)", 2, 6,
	"CSP", 2, 8,
	"ORA <%b,s", 2, 4,
	"TSB <%b", 2, 5,
	"ORA <%b", 2, 3,
	"SHL <%b", 2, 5,
	"ORA [<%b]", 2, 6,
	// 0x8
	"PSH P", 1, 3, // PHP
	"ORA %#", 2, 2,
	"SHL A", 1, 2,
	"PSH D", 1, 4, // PHD Push Direct
	"TSB %w", 3, 6,
	"ORA %w", 3, 4,
	"SHL %w", 3, 6,
	"ORA %L", 4, 5,
	// 0x10
	"BPL @%b", 2, 2,
	"ORA (<%b),Y", 2, 5,
	"ORA (<%b)", 2, 5,
	"ORA (<%b,S),Y", 2, 7,
	"TRB <%b", 2, 5,
	"ORA <%b,X", 2, 4,
	"SHL <%b,X",2, 6,
	"ORA [<%b],Y", 2, 6,
	// 0x18
	"CLC", 1, 2,
	"ORA %w,Y", 3, 4,
	"INC A", 1, 2,
	"TAS", 1, 2,
	"TRB %w", 3, 6,
	"ORA %w,X", 3, 4,
	"SHL %w,X", 3, 7,
	"ORA %L,X", 4, 5,
	// 0x20
	"JSR %w", 3, 6,
	"AND (<%b,X)", 2, 6,
	"JSL %L", 4, 8,
	"AND <%b,s", 2, 4,
	"BIT <%b", 2, 3,
	"AND <%b", 2, 3,
	"ROL <%b", 2, 5,
	"AND [<%b]", 2, 6,
	// 0x28
	"PUL P", 1, 4, // PLP
	"AND %#", 2, 2,
	"ROL A", 1, 2,
	"PUL D", 1, 5, // PLD Pull Direct
	"BIT %w", 3, 4,
	"AND %w", 3, 4,
	"ROL %w", 3, 6,
	"AND %L", 4, 5,
	// 0x30
	"BMI @%b", 2, 2,
	"AND (<%b),Y", 2, 5,
	"AND (<%b)", 2, 5,
	"AND (<%b,s),Y", 2, 7,
	"BIT <%b,X", 2, 4,
	"AND <%b,X", 2, 4,
	"ROL <%b,X", 2, 6,
	"AND [<%b],Y", 2, 6,
	// 0x38
	"SEC", 1, 2,
	"AND %w,Y", 3, 4,
	"DEC A", 1, 2,
	"TSA", 1, 2,
	"BIT %w,X", 3, 4,
	"AND %w,X", 3, 4,
	"ROL %w,X", 3, 7,
	"AND %L,X", 4, 5,
	// 0x40
	"RTI", 1, 7,
	"XOR (<%b,X)", 2, 6,
	"???", 2, 2,
	"XOR <%b,s", 2, 4,
	"MVP XYA %w", 3, 7,
	"XOR <%b", 2, 3,
	"SHR <%b", 2, 5,
	"XOR [<%b]", 2, 6,
	// 0x48
	"PSH A", 1, 3, // PHA
	"XOR %#", 2, 2,
	"SHR A", 1, 2,
	"PSH PBR", 1, 3, // PHK
	"JMP %w", 3, 3,
	"XOR %w", 3, 4,
	"SHR %w", 3, 6,
	"XOR %L", 4, 5,
	// 0x50
	"BVC @%b", 2, 2,
	"XOR (<%b),Y", 2, 5,
	"XOR (<%b)", 2, 5,
	"XOR (<%b,s),Y", 2, 7,
	"MVN XYA %w", 3, 7,
	"XOR <%b,X", 2, 4,
	"SHR <%b,X", 2, 6,
	"XOR [<%b],Y", 2, 6,
	// 0x58
	"CLI", 1, 2,
	"XOR %w,Y", 3, 4,
	"PSH Y", 1, 3, // PHY
	"TAD", 1, 2,
	"JML %L", 4, 4,
	"XOR %w,X", 3, 4,
	"SHR %w,X", 3, 7,
	"XOR %L,X", 4, 5,
	// 0x60
	"RTS", 1, 6,
	"ADC (<%b,X)", 2, 6,
	"PSH @%w", 3, 6, // PER
	"ADC <%b,s", 2, 4,
	"CLR <%b", 2, 3,
	"ADC <%b", 2, 3,
	"ROR <%b", 2, 5,
	"ADC [<%b]", 2, 6,
	// 0x68
	"PUL A", 1, 4, // PLA
	"ADC %#", 2, 2,
	"ROR A", 1, 2,
	"RTL", 1, 6,
	"JMP (%w)", 3, 5,
	"ADC %w", 3, 4,
	"ROR %w", 3, 6,
	"ADC %L", 4, 5,
	// 0x70
	"BVS @%b", 2, 2,
	"ADC (<%b),Y", 2, 5,
	"ADC (<%b)", 2, 5,
	"ADC (<%b,s),Y", 2, 7,
	"CLR <%b,X", 2, 4,
	"ADC <%b,X", 2, 4,
	"ROR <%b,X", 2, 6,
	"ADC [<%b],Y", 2, 6,
	// 0x78
	"SEI", 1, 2,
	"ADC %w,Y", 3, 4,
	"PUL Y", 1, 4, // PLY
	"TDA", 1, 2,
	"JMP (%w,X)", 3, 6,
	"ADC %w,X", 3,4,
	"ROR %w,X", 3,7,
	"ADC %L,X",4,5,
	// 0x80
	"BRA @%b", 2, 2,
	"STA (<%b,X)", 2, 6,
	"BRL @%w", 3, 3,
	"STA <%b,S", 2, 4,
	"STY <%b", 2, 3,
	"STA <%b", 2, 3,
	"STX <%b", 2, 3,
	"STA [<%b]", 2, 6,
	// 0x88
	"DEY", 1, 2,
	"BIT %#", 2, 2,
	"TXA", 1, 2,
	"PSH DBR", 1, 3, // PHB
	"STY %w", 3, 4,
	"STA %w", 3, 4,
	"STX %w", 3, 4,
	"STA %L", 4, 5,
	// 0x90
	"BCC @%b", 2, 2,
	"STA (<%b),Y", 2, 6,
	"STA (<%b)", 2, 5,
	"STA (<%b,s),Y", 2, 7,
	"STY <%b,X", 2, 4,
	"STA <%b,X", 2, 4,
	"STX <%b,Y", 2, 4,
	"STA [<%b],Y", 2, 6,
	// 0x98
	"TYA", 1, 2,
	"STA %w,Y", 3, 5,
	"TXS", 1, 2,
	"TXY", 1, 2,
	"CLR %w", 3, 4,
	"STA %w,X", 3, 5,
	"CLR %w,X", 3, 5,
	"STA %L,X", 4, 5,
	// 0xA0
	"LDY %#i", 2, 2,
	"LDA (<%b,X)", 2, 6,
	"LDX %#i", 2, 2,
	"LDA <%b,S", 2, 4,
	"LDY <%b", 2, 3,
	"LDA <%b", 2, 3,
	"LDX <%b", 2, 3,
	"LDA [<%b]", 2, 6,
	// 0xA8
	"TAY", 1, 2,
	"LDA %#", 2, 2,
	"TAX", 1, 2,
	"PUL DBR", 1, 4, // PLB
	"LDY %w", 3, 4,
	"LDA %w", 3, 4,
	"LDX %w", 3, 4,
	"LDA %L", 4, 5,
	// 0xB0
	"BCS @%b", 2, 2,
	"LDA (<%b),Y", 2, 5,
	"LDA (<%b)", 2, 5,
	"LDA (<%b,S),Y", 2, 7,
	"LDY <%b,X", 2, 4,
	"LDA <%b,X", 2, 4,
	"LDX <%b,Y", 2, 4,
	"LDA [<%b],Y", 2, 6,
	// 0xB8
	"CLV", 1, 2,
	"LDA %w,Y", 3, 4,
	"TSX", 1, 2,
	"TYX", 1, 2,
	"LDY %w,X", 3, 4,
	"LDA %w,X", 3, 4,
	"LDX %w,Y", 3, 4,
	"LDA %L,X",4, 5,
	// 0xC0
	"CPY %#i", 2, 2,
	"CMP (<%b,X)", 2, 6,
	"CLP %#b", 2, 3,
	"CMP <%b,S", 2, 4,
	"CPY <%b", 2, 3,
	"CMP <%b", 2, 3,
	"DEC <%b", 2, 5,
	"CMP [<%b]", 2,6,
	// 0xC8
	"INY", 1, 2,
	"CMP %#", 2, 2,
	"DEX", 1, 2,
	"WAI", 1, 3,
	"CPY %w", 3, 4,
	"CMP %w", 3, 4,
	"DEC %w", 3, 6,
	"CMP %L", 4, 5,
	// 0xD0
	"BNE @%b", 2, 2,
	"CMP (<%b),Y", 2, 5,
	"CMP (<%b)", 2, 5,
	"CMP (<%b,S),Y", 2, 7,
	"PSH %b", 2, 6, // PEI
	"CMP <%b,X", 2, 4,
	"DEC <%b,X", 2, 6,
	"CMP [<%b],Y", 2, 6,
	// 0xD8
	"CLD", 1, 2,
	"CMP %w,Y", 3, 4,
	"PSH X", 1, 3, // PHX
	"HLT", 1, 3, // STP
	"JML (%w)", 3, 6,
	"CMP %w,X", 3, 4,
	"DEC %w,X", 3, 7,
	"CMP %L,X", 4, 5,
	// 0xE0
	"CPX %#i", 2, 2,
	"SBC (<%b,X)", 2, 6,
	"SEP %#b", 2, 3,
	"SBC <%b,s", 2, 4,
	"CPX <%b", 2, 3,
	"SBC <%b", 2, 3,
	"INC <%b", 2, 5,
	"SBC [<%b]", 2, 6,
	// 0xE8
	"INX", 1, 2,
	"SBC %#", 2, 2,
	"NOP", 1, 2,
	"SWA", 1, 3,
	"CPX %w", 3, 4,
	"SBC %w", 3, 4,
	"INC %w", 3, 6,
	"SBC %L", 4, 5,
	// 0xF0
	"BEQ @%b", 2, 2,
	"SBC (<%b),Y", 2, 5,
	"SBC (<%b)", 2, 5,
	"SBC (<%b,S),Y", 2, 7,
	"PSH %#w", 3, 5, // PEA
	"SBC <%b,X", 2, 4,
	"INC <%b,X", 2, 6,
	"SBC [<%b],Y", 2, 6,
	// 0xF8
	"SED", 1, 2,
	"SBC %w,Y", 3, 4,
	"PUL X", 1, 4, // PLX
	"XCE", 1, 2,
	"JSR (%w,X)", 3, 6,
	"SBC %w,X", 3, 4,
	"INC %w,X", 3, 7,
	"SBC %L,X", 4, 5
};

// ADDRESSING MODES - several categories, one mode from each category can be combined
// None of the instructions can use all combinations, and certain combinations are
// never used with any instruction.
	// Used to calculate memory address:
	//   Absolute: Argument is two bytes and specifies a memory location
	//     + Long: Argument is three bytes instead             opcode $addr
	//   Direct: Argument is one byte; memory address formed by adding D.
	//     Bank is always 0, unless indirect addressing is also used, in
	//     which case the bank is the program bank.            opcode <$addr
	//   Stack Relative: Argument is one byte, which is added to the stack pointer
	//     to find the final address.                          opcode $addr,S
	// Indexed addressing:
	//   Indexed: Memory location argument is added to X or Y to get the final
	//     memory location.                                    opcode $addr,X or Y
	//   Note that certain indexed addressing modes only work with the X register
	//     or only with the Y register.
	//
	// Note! The above addressing mode specifiers are used only to specify the
	// size of the argument and how it is used.  The specifiers DO NOT SPECIFY
	// anything about the actual memory location being referenced.
	// For example, suppose the Absolute Indexed mode is used.  This means that
	// the argument to the opcode is 16 bits and will be added to an index
	// register.  It does not, however, mean that the value at the calculated
	// memory location will be a 16-bit value.  The size of the value at the memory
	// location is determined by either:
	//   a. whether the processor is in 16-bit mode (and whether E=0), or
	//   b. if indirect addressing is used, whether the Long Indirect mode is used.
	//
	// Indirect addressing:
	//   Indirect: Uses indirection (pointers).  In other words, after the memory
	//     location is found and the contents are loaded, the contents are used as
	//     a SECOND memory location and the value at the second memory location is
	//     used in the operation.                              opcode ($addr)
	//     + Direct Long: Normally, the pointer found at the specified location
	//       will be two bytes.  If Direct Indirect Long is used, the pointer
	//       at the memory address found will be a three-byte long address.
	//                                                         opcode [$addr]
	// Indexed Indirect (preindexed; (arg,x)) vs. Indirect Indexed (postindexed;
	// (arg),x): remember, these two are different!
// There is also two more modes, which are never used with any other specifiers:
	// Implied: Opcode has no arguments (The "accumulator" addressing mode is like
	//   an implied mode and it means it acts on the accumulator only.)
	// Immediate: The argument specifies an actual constant value to be used,
	//   rather than a memory address. [opcode #$XX or #$XXXX]
	//   some immediate addressing modes always have a 8-bit argument (CLP, SEP);
	//   others have an 8/16 bit argument depending on the status of the M bit;
	//   others have an 8/16 bit argument depending on the status of the X bit.
	//   (When the programmer writes a immediate addressing instruction he must make
	//   sure the assembler knows what size to make it!)

#define AA_ROM (byte *)0x1
#define AA_HIROM (byte *)0x2
//#define AA_LOROM_x40 (byte *)0x3
#define AA_NOTHING (byte *)0x4
	// Placeholder in the addressarea array, needed since the rom variable is initialized to NULL

/*struct addressarea { // old mem map based on Duncanthrax's doc
	// this is a representation of the SNES memory map. It's used by the getstartaddresses()
	// function to build the address mapping table, *startaddr[].
	byte b1, b2;
	word a1, a2;
	byte *start;
	boolean continuous;
} loromarea [] = {
	{ 0x00, 0x3F, 0x0000, 0x1FFF, ram,             },
	{ 0x00, 0x3F, 0x2000, 0x5FFF, registers,       },
	{ 0x00, 0x3F, 0x6000, 0x7FFF, expram,          },
	{ 0x00, 0x3F, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0x40, 0x7C, 0x0000, 0x7FFF, AA_ROM,    true  },
	{ 0x40, 0x7C, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0x7D, 0x7D, 0x0000, 0xFFFF, sram,            },
	{ 0x7E, 0x7F, 0x0000, 0xFFFF, ram,       true  },

	{ 0x80, 0xBF, 0x0000, 0x1FFF, ram,             },
	{ 0x80, 0xBF, 0x2000, 0x5FFF, registers,       },
	{ 0x80, 0xBF, 0x6000, 0x7FFF, expram,          },
	{ 0x80, 0xBF, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0xC0, 0xFC, 0x0000, 0x7FFF, AA_ROM,    true  },
	{ 0xC0, 0xFC, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0xFD, 0xFD, 0x0000, 0xFFFF, sram,            },
	{ 0xFE, 0xFF, 0x0000, 0xFFFF, ram,       true  },
	{ 0 }
}, hiromarea [] = {
	{ 0x00, 0x3F, 0x0000, 0x1FFF, ram,             },
	{ 0x00, 0x3F, 0x2000, 0x5FFF, registers,       },
	{ 0x00, 0x3F, 0x6000, 0x7FFF, sram,            },
	{ 0x00, 0x3F, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0x40, 0x7D, 0x0000, 0xFFFF, AA_ROM,    true  },
	{ 0x7E, 0x7F, 0x0000, 0xFFFF, ram,       true  },

	{ 0x80, 0xBF, 0x0000, 0x1FFF, ram,             },
	{ 0x80, 0xBF, 0x2000, 0x5FFF, registers,       },
	{ 0x80, 0xBF, 0x6000, 0x7FFF, sram,            },
	{ 0x80, 0xBF, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0xC0, 0xFF, 0x0000, 0xFFFF, AA_ROM,    true  },
	{ 0 }
};*/

struct addressarea {
	// this is a representation of the SNES memory map. It's used by the getstartaddresses()
	// function to build the address mapping table, *startaddr[].
	byte b1, b2;
	word a1, a2;
	byte *start;
	boolean continuous;
	word spacing;
} loromarea [] = {
	{ 0x00, 0x3F, 0x0000, 0x1FFF, ram,             },
	{ 0x00, 0x3F, 0x2000, 0x5FFF, registers,       },
	{ 0x00, 0x3F, 0x6000, 0x7FFF, expram,          },
	{ 0x00, 0x6F, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0x40, 0x6F, 0x0000, 0x7FFF, AA_NOTHING       },
	{ 0x70, 0x77, 0x0000, 0xFFFF, sram             },
	{ 0x78, 0x7D, 0x0000, 0xFFFF, AA_NOTHING       },
	{ 0x7E, 0x7F, 0x0000, 0xFFFF, ram,       true  },

	{ 0x80, 0xBF, 0x0000, 0x1FFF, ram,             },
	{ 0x80, 0xBF, 0x2000, 0x5FFF, registers,       },
	{ 0x80, 0xBF, 0x6000, 0x7FFF, expram,          },
	{ 0x80, 0xEF, 0x8000, 0xFFFF, AA_ROM,    true  },
	{ 0xC0, 0xEF, 0x0000, 0x7FFF, AA_NOTHING       },
	{ 0xF0, 0xF7, 0x0000, 0xFFFF, sram             },
	{ 0xF8, 0xFD, 0x0000, 0xFFFF, AA_NOTHING       },
	{ 0xFE, 0xFF, 0x0000, 0xFFFF, ram,       true  },
	{ 0 }
}, hiromarea [] = {
	{ 0x00, 0x3F, 0x0000, 0x1FFF, ram,             },
	{ 0x00, 0x3F, 0x2000, 0x5FFF, registers,       },
	{ 0x00, 0x2F, 0x6000, 0x7FFF, expram,          },
	{ 0x30, 0x3F, 0x6000, 0x7FFF, sram,      true  },
	{ 0x00, 0x3F, 0x8000, 0xFFFF, AA_HIROM,  true, 0x8000 },
	{ 0x40, 0x6F, 0x0000, 0xFFFF, AA_ROM,    true  },
	{ 0x70, 0x77, 0x0000, 0xFFFF, sram             },
	{ 0x78, 0x7D, 0x0000, 0xFFFF, AA_NOTHING       },
	{ 0x7E, 0x7F, 0x0000, 0xFFFF, ram,       true  },

	{ 0x80, 0xBF, 0x0000, 0x1FFF, ram,             },
	{ 0x80, 0xBF, 0x2000, 0x5FFF, registers,       },
	{ 0x80, 0xAF, 0x6000, 0x7FFF, expram,          },
	{ 0xB0, 0xBF, 0x6000, 0x7FFF, sram,      true  },
	{ 0x80, 0xBF, 0x8000, 0xFFFF, AA_HIROM,  true, 0x8000 },
	{ 0xC0, 0xFF, 0x0000, 0xFFFF, AA_ROM,    true  },
	{ 0 }
};

void getstartaddresses (boolean hirom)
{
	int x, i;
	int b, a; // bank, address
	struct addressarea *aa;

	if (hirom) aa = hiromarea;
	else aa = loromarea;
	for (x = 0; aa[x].start != NULL; x++) {
		if (aa[x].start == AA_ROM)
			aa[x].start = rom;
		if (aa[x].start == AA_HIROM)
			aa[x].start = rom + 0x8000;
		//if (aa[x].start == AA_ROM_x40)
		//	aa[x].start = rom + 0x40*0x8000;
		if (aa[x].start == AA_NOTHING)
			aa[x].start = expram;
	}
	for (i = 0; i <= 0x7FF; i++) {
			// Get bank and address
		b = (i >> 3);
		a = (i & 7) << 13;
			// Find which of the address areas that bb.aaaa is part of
		for (x = 0; aa[x].start != NULL; x++) {
			if (b >= aa[x].b1 && b <= aa[x].b2 && a >= aa[x].a1 && a <= aa[x].a2)
				break;
		}
		startaddr [i] = (aa[x].start + (a - aa[x].a1));
		if (aa[x].continuous)
			startaddr [i] += (b - aa[x].b1) * (aa[x].a2 - aa[x].a1 + 1 + aa[x].spacing);
		assert (aa[x].start != NULL);
			// Map that address to the array specified by start.
		//debug0 ("Mapping memory: %X.%X->%P[%X]", b, a, aa[x].start, startaddr [i]-aa[x].start);
	}
}

void resetsystem (boolean warmreset)
{
	int x;

	if (!romloaded) return;
	getstartaddresses (ishirom());
	if (warmreset) {
		//reg.S = (reg.S & 0xFF) | 0x100;
		reg.S = 0x1FF;
		//*(byte*)SNESMEM(reg.S) = (byte)(reg.PC >> 16); (*(word*)&reg.S)--;
		//*(word*)SNESMEM(reg.S-1) = (word)reg.PC; (*(word*)&reg.S) -= 2;
		//*(byte*)SNESMEM(reg.S) = (byte)(reg.P); (*(word*)&reg.S)--;
	} else {
		reg.S = 0x1FF;
		dealloc_autobacktrack();
		if (autobacktrack)
			alloc_autobacktrack(backtracknum);
	}
	reg.D_dummy = 0x0000;
	reg.P_dummy = 0x34; //|= 0x34; // 0011 0100
	//reg.P &= ~BCD;
	reg.A_dummy = 0;
	reg.X_dummy = 0; //&= 0xFF;
	reg.Y_dummy = 0; //&= 0xFF;
	reg.E_dummy = true;
	reg.PC = curheader.resetvector; //| (curheader.fastrom ? 0x800000 : 0);
	debug0 ("reg.PC = %LX", reg.PC);
	reg.DBR_dummy1 = reg.DBR_dummy2 = reg.DBR = reg.DBR_dummy3 = 0;
	romrunning = false;
	total_frames = 0;
	debug_instr = 0;

	memset (ram, 0, sizeof (ram));
	memset (registers, 0, sizeof (registers));
	memset (expram, 0, sizeof (expram));
	memset (vram, 0, sizeof (vram));
	memset (oam, 0, sizeof (oam));
	memset (cgram, 0, sizeof (cgram));
	for (x = 0; x < 256; x++) {
		((word*)cgram)[x] = ((custompalette[x] >> 19) & 31) + (((custompalette[x] >> 11) & 31) << 5)\
		 + (((custompalette[x] >> 3) & 31) << 10);  // Initialize to custom palette
	}
	memset (&state, 0, sizeof (ppuinternalregisters));
	state.hcounter = state.vcounter = 0;
	state.reg4210 = state.reg4211 = 1;
	state.vramincrate = 2;
	memset (instr_tally, 0, sizeof (instr_tally));
	memset (mmio_rtally, 0, sizeof (mmio_rtally));
	memset (mmio_wtally, 0, sizeof (mmio_wtally));
    memset (cstatus, 0, sizeof (cstatus));  // Recaching required
	memset (spr8x8size, 0, sizeof (spr8x8size));
	frame_cycles = total_frames = -1; // Force start of new frame
}

char *getaccessoryname (char *fn, char *dir, char *ext)
{
	static char full [100], full2 [100];
	int p;

	debug0 ("Get accessory name: fn %s dir %s ext %s", fn, dir, ext);
	full [0] = '\0';
	getabspath (full);
	combinepaths (full2, full, dir);
	combinepathfile (full, full2, fn);
	p = findlastch (full, '.') + 1;
	if (p == 0) {
		p = strlen (full) + 1;
		full [p - 1] = '.';
	}
	strcpy (full + p, ext);
	debug0 ("  Got path: %s", full);
	return full;
}

void savesram ()
{
	char *fn = getaccessoryname (curgs->filename, sramdir, "SRM");
	int size;
	FILE *fp;

	for (size = 0xFFFF; size > 0; size--)
		if (sram[size] != 255)
			break;
	if (size == 0) {
		debug0 ("Not saving SRAM %s because size is zero.", fn);
		return;
	}
	if (vidmode == 0) printf ("Saving SRAM... (%s)", fn);
	else              addmessage ("Saving SRAM... (%s)", fn);
		// now round up
	if (size > 32768) size = 65536;
	else if (size > 16384) size = 32768;
	else if (size > 8192) size = 16384;
	else if (size > 4096) size = 8192;
	else if (size > 2048) size = 4096;
	else if (size > 1024) size = 2048;
	else size = 1024;
	if ((fp = fopen (fn, "wb")) == NULL) {
		if (vidmode == 0) printf ("Error opening SRAM file to save!");
		else              addmessage ("Error opening SRAM file to save!");
		return;
	}
	fwrite (sram, size, 1, fp);
	fclose (fp);
}

void loadsram ()
{
	char *fn = getaccessoryname (curgs->filename, sramdir, "SRM");
	FILE *fp;

	memset (sram, 255, 0x10000);
	if (vidmode == 0) printf ("Loading SRAM (%s)...", fn);
	else              addmessage ("Loading SRAM (%s)...", fn);
	if ((fp = fopen (fn, "rb")) == NULL) {
		if (vidmode == 0) printf ("Error opening SRAM file to load!");
		else              addmessage ("Error opening SRAM file to load!");
		return;
	}
	fread (sram, 1, 0x10000, fp);
	fclose (fp);
}

int cursaveslot = 0;
	// supports 10 save slots (0-9)

// quick save ID's
#define QSRAM 1
#define QSREGISTERS 2
#define QSSRAM 3
#define QSEXPRAM 4
#define QSCGRAM 5
#define QSOAM 6
#define QSSIXVARS 7 /* frame_cycles, screen_scanline */
#define QSSTATE 8
#define QSVRAM 9
#define QSMMIO2100 10 /* $2100 to $21FF */
#define QSMMIO4200 11 /* $4200 to $43FF */

boolean instantsave (int slot, boolean autobacktrack)
{   // Returns OKAY if save worked
	char *fn, ext[4] = "QSx"; // "Quick Save X"
	FILE *fp;
	char fileid[4] = "QS\x1A";
	int id, size;
	char s[80];

	ext [2] = slot + '0';
	if (autobacktrack) {
		s[0] = slot + '0';
		s[1] = '\0'; fn = s;
		addmessage ("[Checkpoint]");
	} else {
		fn = getaccessoryname (curgs->filename, instdir, ext);
	}
	if ((fp = vfopen (fn, "wb")) == NULL) {
		addmessage ("Couldn't save state %s...", fn);
		return -1;
	}
	vfwrite (fileid, 1, 4, fp);
	id = QSREGISTERS; size = sizeof (regpack);
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (&reg, size, 1, fp);
	id = QSSTATE; size = sizeof (ppuinternalregisters);
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (&state, size, 1, fp);
	id = QSCGRAM; size = 256*2;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (cgram, size, 1, fp);
	id = QSOAM; size = 544;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (oam, size, 1, fp);
	id = QSSIXVARS; size = sizeof(int) * 6;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (&frame_cycles,    sizeof(int), 1, fp);
	vfwrite (&scan_cycles,     sizeof(int), 1, fp);
	vfwrite (&screen_scanline, sizeof(int), 1, fp);
	vfwrite (&cycles_per_scan, sizeof(int), 1, fp);
	vfwrite (&total_frames,    sizeof(int), 1, fp); // Optional
	vfwrite (&debug_instr,     sizeof(int), 1, fp); // Optional
	id = QSMMIO2100; size = 0xFF;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (registers + 0x0100, size, 1, fp);
	id = QSMMIO4200; size = 0x1FF;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (registers + 0x2200, size, 1, fp);
	id = QSRAM; size = 128*1024;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (ram, size, 1, fp);
	id = QSVRAM; size = 0x10000;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (vram, size, 1, fp);
	id = QSEXPRAM; size = 0x2000;
	vfwrite (&id, sizeof(int), 1, fp);
	vfwrite (&size, sizeof(int), 1, fp);
	vfwrite (expram, size, 1, fp);
	vfclose (fp);
	if (fn[1] != '\0') {
		sprintf (s, "State saved to %s.", fn);
		if (debugmode)
			adddebugline (s, false, '!');
		else
			addmessage ("%s", s);
	}
	return OKAY;
}

boolean instantload (int slot, boolean autobacktrack)
	// Returns OKAY if load worked
{
	char *fn, ext[4] = "QSx"; // ".Quick Save X"
	FILE *fp;
	char fileid[4];
	dword id, size;
	int x;
	char s[80];

	ext [2] = slot + '0';
	if (autobacktrack) {
		s[0] = slot + '0';
		s[1] = '\0'; fn = s;
	} else {
		fn = getaccessoryname (curgs->filename, instdir, ext);
	}
	if ((fp = vfopen (fn, "rb")) == NULL) {
		addmessage ("Couldn't open state %s", fn);
		return -1;
	}
	debug0 ("got here");
	vfread (fileid, 1, 4, fp);
	if (fileid[0]!='Q' || fileid[1]!='S' || fileid[2]!=26 || fileid[3]!=0) {
		addmessage ("%s: unrecognised file format!", fn);
		fclose (fp);
		return -1;
	}
	resetsystem (true);
	resetscreenframe ();
	applycustompalette();
	while (!vfeof(fp)) {
		vfread (&id, 1, 4, fp);
		vfread (&size, 1, 4, fp);
		switch (id) {
		case QSRAM:
			vfread (ram, size, 1, fp);
			break;
		case QSREGISTERS:
			vfread (&reg, size, 1, fp);
			break;
		case QSSRAM:
			memset (sram, 0xFF, 0x10000); // Clear SRAM
			vfread (sram, size, 1, fp);
			break;
		case QSEXPRAM:
			vfread (expram, size, 1, fp);
			break;
		case QSCGRAM:
			vfread (cgram, size, 1, fp);
			break;
		case QSOAM:
			vfread (oam, size, 1, fp);
			break;
		case QSSIXVARS:
			vfread (&frame_cycles, sizeof(int), 1, fp);
			vfread (&scan_cycles, sizeof(int), 1, fp);
			vfread (&screen_scanline, sizeof(int), 1, fp);
			vfread (&cycles_per_scan, sizeof(int), 1, fp);
			if (size > 16) vfread (&total_frames, sizeof(int), 1, fp);
			if (size > 20) vfread (&debug_instr, sizeof(int), 1, fp);
			for (x = 24; x < size; x++) // ignore any extra data
				fgetc (fp);
			break;
		case QSSTATE:
			if (size > sizeof (ppuinternalregisters)) {
				vfread (&state, sizeof (ppuinternalregisters), 1, fp);
				for (x = sizeof (ppuinternalregisters); x < size; x++)
					fgetc (fp); // ignore any extra data
			} else {
				vfread (&state, size, 1, fp);
			}
			break;
		case QSVRAM:
			vfread (vram, size, 1, fp);
			break;
		case QSMMIO2100:
			vfread (registers + 0x0100, size, 1, fp);
			break;
		case QSMMIO4200:
			vfread (registers + 0x2200, size, 1, fp);
			break;
		default:
			for (x = 0; x < size; x++)
				fgetc (fp); // ignore the data
			break;
		}
	}
	vfclose (fp);
	if (autobacktrack)
		sprintf (s, "[Backtracked to last checkpoint.]");
	else
		sprintf (s, "%s Loaded.", fn);
	if (debugmode)
		adddebugline (s, false, '!');
	else
		addmessage ("%s", s);
	romrunning = true;
	return OKAY;
}

void unassemble (dword opdata, char *s, byte P)
{
	byte opcode = (byte) opdata;
	char *format = opcodelist[opcode].format;
	boolean isword = false, islong = false, issigned = false, isnumber = false;
	int n, x;

	opdata >>= 8;
	for (;;) {
		if (*format == '%') {
			format++;
			switch (*format) {
			case '#':
				format++;
				isnumber = true;
				switch (*format) {
				case 'i':
					if (!(P & INDEX)) isword = true;
					break;
				case 'b':
					break;
				case 'w':
					isword = true;
					break;
				default:
					format--; // literal--to be printed
					if (!(P & MEMORY)) isword = true; break;
				}
				break;
			case 'b':
				break;
			case 'w':
				isword = true;
				break;
			case 'L':
				islong = true;
				break;
			}
			if (isnumber)
				*s++ = '#';
			if (issigned) {
				if (isword) n = (signed short) opdata;
				if (islong) n = opdata;
				else n = (signed char) opdata;
				x = sprintf (s, "%d", n);
			} else {
				if (isword) {
					n = (unsigned short) opdata;
					x = sprintf (s, "$%0.4X", n);
				} else  if (islong) {
					n = opdata;
					x = sprintf (s, "$%0.6X", n);
				} else {
					n = (unsigned char) opdata;
					x = sprintf (s, "$%0.2X", n);
				}
			}
			s += x;
			format++;
		} else {
			*s = *format;
			if (*s == '@') issigned = true;
			if (*s == '\0') break;
			s++;
			format++;
		}
	}
}

char *getdescr (byte opcode)
{
	char *opc = opcodelist[opcode].format;
	int x;

	for (x = 0; instrlist[x].mnemonic[0] != '\0'; x++) {
		if (instrlist[x].mnemonic[0] == opc[0] &&
			instrlist[x].mnemonic[1] == opc[1] &&
			instrlist[x].mnemonic[2] == opc[2]) {
			return instrlist[x].descr;
		}
	}
	return "[Search error: couldn't find opcode decription]";
}

boolean ispal ()
{
	if (curgs->forcecountry == FORCEPAL) return true;
	else if (curgs->forcecountry == FORCENTSC || curheader.country < 2) return false;
	else return true;
}

boolean ishirom ()
{
	if (curgs->forcerom == FORCEHIROM) return true;
	else if (curgs->forcerom == FORCELOROM || curheader.hirom == false) return false;
	else return true;
}

int curfastrom;
long framestarttime;
#define FTHISTORY 30
byte frametime [FTHISTORY];
boolean framewasrendered [FTHISTORY];

void getfps (int &vfps, int &rfps)
{   // returns the fps * 10. (One decimal place included..)
	int x, rf, t;

	x = 0, rf=0, t=0;
	do {
		t += frametime [x];
		if (framewasrendered[x]) rf++;
		x++;
	} while (x < FTHISTORY);
	if (t != 0) {
		vfps = FTHISTORY * 10000 / t;
		rfps = rf * 10000 / t;
	} else {
		vfps = rfps = 0;
	}
	setcolor (0);
}

#define HBCOMBINE(a) ((*REG4302(x) & 0xFF0000) | (a))
// Combine offset with HDMA Bank

inline int getcount (int x)
{
	int temp;

	temp = HBCOMBINE(*(word*)REG4308(x));
	temp = *SNESMEM(temp) & 0x7F;
	//if (temp > 0x80) temp &= 0x7F;
	if (temp == 0) state.hdma_inprogress &= ~(1 << x);
	return temp;
}

void resethdma ()
{
	int x;

	state.hdma_inprogress = *REG420C;
	for (x = 0; x < 8; x++) {
		if (!(*REG420C & (1 << x)))
			continue;
		if (!curgs->indirecthdma && (*REG4300(x) & 0x40)) { // Indirect
			debug0 ("Ignoring HDMA Indirect channel %d", x);
			state.hdma_inprogress &= ~(1 << x);
		}

		if (*REG4300(x) & 0x40) { // Indirect
			*REG4308(x) = *(word*)REG4302(x);
			// ^ Pointer to pre-table
			*(word*)REG4305(x) = *(word*)SNESMEM ((*(dword*)REG4302(x) & 0xFFFFFF) + 1);
			//*((byte*)REG4305(x) + 2) = *((byte*)REG4302(x) + 2); // Bank
			//Apparently the bank should be set in regular HDMA and not indirect HDMA
			// ^ Pointer to final table (data values)
		} else { // Regular HDMA
			*(word*)REG4305(x) = *(word*)REG4302(x); // Not used in regular HDMA?  Or perhaps used for continuous mode.  Yes, that's it.
			*((byte*)REG4305(x) + 2) = *((byte*)REG4302(x) + 2); // Bank
			*REG4308(x) = *(word*)REG4302(x); // Pointer to table
		}
		*REG430A(x) = getcount (x);

		//*REG430A(x) = *SNESMEM(*REG4302(x) & 0xFFFFFF) - 1;
			// Get count byte, first byte of table
		debug0 ("HDMA Channel %d mode $%X cpu$%X reg$%X waitcount%d", x, *REG4300(x), *REG4302(x) & 0xFFFFFF, *REG4301(x) | 0x2100, *REG430A(x));
		//tmp = SNESMEM (*REG4302(x) & 0xFFFFFF);
		//for (y=0; y < 3; y++) {
		//	debug0 ("Table: %0.2X %0.2X %0.2X %0.2X %0.2X %0.2X %0.2X %0.2X %0.2X %0.2X", tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5], tmp[6], tmp[7], tmp[8], tmp[9]);
		//	tmp += 10;
		//}
	}
	debug0 ("Inprogress this frame: %0.2X", state.hdma_inprogress);
	//dohdmaline(); // First line
}

inline int dmadatasize (int mode)
{
	static int modes[] = { 1,2,2,4,4,0,0,0 };
	return modes [mode & 7];
}

int writehdma (int mode, int addr, int breg)
{	// Returns size of data written.
	int size, regand;
	byte *cpu, *reg;
	int y;

	switch (mode & 0x7) {
	case 0: size = 1; regand = 0; break;
	case 1: size = 2; regand = 1; break;
	case 2: size = 2; regand = 0; break;
	case 3: size = 4; regand = 2; break;
	case 4: size = 4; regand = 3; break;
	default:size = 0; regand = 0; debug0 ("Invalid HDMA Mode %d b$%X",mode,breg); break;
	}
	//debug0 ("HDMA sc%d mem%X %X %X %X ", screen_scanline, *SNESMEM(addr), *SNESMEM(addr+1), *SNESMEM(addr+2), *SNESMEM(addr+3));
	for (y = 0; y < size; y++) {
		cpu = SNESMEM (addr + y); // Value to write
		reg = SNESMEM (0x2100 | (breg + (regand == 2 ? (y >> 1) : (y & regand))));
		*reg = *cpu;
		trapregwrite (reg, false);
	}
	return size;
}

void getpointers (int x, int &addr, int &iaddr)
{
	if (!(*REG4300(x) & 0x40)) {
		addr = HBCOMBINE (*REG4308(x));
		// addr is SNES address of HDMA table at current offset
		iaddr = (*(dword*)REG4305(x) + 1) & 0xFFFFFF; // Equal to addr+1 when NOT in continuous mode.
		// iaddr is address of the actual value(s) to write
	} else {
		// REG4305,6,7 used as address after indirection; 4308,9 used before.
		addr = HBCOMBINE (*REG4308(x));
		// addr is SNES address of HDMA pre-table at current offset
		iaddr = *(dword*)REG4305(x) & 0xFFFFFF; // Equal to addr+1 when NOT in continuous mode.
		// iaddr is address of the actual value(s) to write
	}
}

void dohdmaline ()
{
	int x;
	int addr, iaddr, size;

	for (x = 0; x < 8; x++) {
		if (!(state.hdma_inprogress & (1 << x)))
			continue;
		if (*REG430A(x) == 0) {
			if (!(*REG4300(x) & 0x40)) {
				(*REG4305(x))++;
				*REG4308(x) = *REG4305(x);
			} else {
				(*REG4308(x)) += 3; // Pre-table entries are always spaced three bytes apart.
				*REG4305(x) = *(word*)SNESMEM(HBCOMBINE(*REG4308(x)+1));
			}
			*REG430A(x) = getcount (x);
			if (!(state.hdma_inprogress & (1 << x)))
				continue;
		}
		getpointers (x, addr, iaddr);
		if (*REG430A(x) == getcount (x) || *SNESMEM(addr) & 0x80) {
			size = writehdma (*REG4300(x), iaddr, *REG4301(x));
			*(word*)REG4305(x) += size;
		}
		(*REG430A(x))--;
	}
}

inline void executecpu ()
{
#ifdef USE_C_CPU_CORE
	switch (reg.P & (MEMORY | INDEX)) {
	case 0:
		emulate_m0_x0 (); return;
	case MEMORY:
		emulate_m1_x0 (); return;
	case INDEX:
		emulate_m0_x1 (); return;
	case MEMORY | INDEX:
		emulate_m1_x1 (); return;
	}
#else
	doemulate ();
#endif
}

void debug_executecpu (int cycles)
{
	int left, temp;

	// frame cycles initted to 0, so cpu frame is always started at the beginning of the frame.
	// If debugger is entered from game, new frame will already be initted.
	if (frame_cycles <= 0) {
		startcpuframe ();
		total_frames++;
		startscanline ();
		debug0 ("Frame %d: debug_instr=%d", total_frames, debug_instr);
	}
	temp = scan_cycles;
	scan_cycles = cycles;
	executecpu ();
	left = scan_cycles;
	scan_cycles = temp + left - cycles;
	if (scan_cycles < 0) {
		frame_cycles -= cycles_per_scan;
		scan_cycles += cycles_per_scan;
		screen_scanline++;
		startscanline ();
		//if (screen_scanline < 10 || screen_scanline > 220)
		//	debug0 ("scanline %d new scan_cycles %d frame_cycles %d PC$%X", screen_scanline, scan_cycles, frame_cycles, reg.PC);
	}
}

boolean debug_nmistart;

void startvblank ()
{
	int x;

	state.reg4210 |= 0x80; // VBL start occurred--Required even if screen is off
	if (!(*REG4200 & 0x80)) {
		return; // NMI Disabled
	}
	if (reg.E) x = *(word*)SNESMEM(0x00FFFA);
	else x = *(word*)SNESMEM(0x00FFEA);
	if (x != 0) {
		if (debugmode) adddebugline ("[Jumped to NMI vector]", false, '*');
		state.end_wai = true;
		*(byte*)SNESMEM(reg.S) = (byte)(reg.PC >> 16); (*(word*)&reg.S)--;
		*(word*)SNESMEM(reg.S-1) = (word)(reg.PC); (*(word*)&reg.S) -= 2;
		*(byte*)SNESMEM(reg.S) = (byte)(reg.P); (*(word*)&reg.S)--;
			// Push PC and P
		reg.P |= 4; // Interrupt flag
		reg.PC = x;
		debug0 ("Jumped to NMI Vector: %LX; E = %d", reg.PC, reg.E);
		debug_nmistart = true;
	}
}

void startcpuframe ()
{
	int hz;

	state.reg4210 &= ~0x80; // NMI over

	// decide on how many cycles to do this frame
	if (curheader.fastrom && (*SNESMEM (0x420D) & 1) && !curgs->ignorefastrom) { // fastrom
		hz = 3580000 * curgs->cpupercent / 100;
		curfastrom = true;
	} else {
		hz = 2680000 * curgs->cpupercent / 100;
		curfastrom = false;
	}
	if (ispal ()) {
		frame_cycles = hz / 50;
	} else {
		frame_cycles = hz / 60;
	}

	cycles_per_scan = scan_cycles = frame_cycles / 260;
	screen_scanline = 0; // start with scanline 0

	// Restart HDMA here
	resethdma ();

	resetscreenframe ();

	setjoypadreg (0);
	setjoypadreg (1);
	setjoypadreg (2);
	setjoypadreg (3);
}

void startscanline ()
{
	int x;

	if (!(reg.P & 0x4) /*&& !(state.reg4211 & 0x80)*/) { // IRQs possible
		//if (((*REG4200 & 0x10) && ((*REG4207 & 0x1FF) <= 339)) || // H-IRQs
		//    ((*REG4200 & 0x20) && ((*REG4209 & 0x1FF) == screen_scanline))) { // V-IRQs
		if ((*REG4200 & 0x30) && (*REG4209 & 0x1FF) == screen_scanline) {
			state.reg4211 |= 0x80; // IRQ start occurred
			if (reg.E) x = *(word*)SNESMEM(0x00FFFE);
			else x = *(word*)SNESMEM(0x00FFEE);
			if (x != 0) {
				state.end_wai = true;
				*(byte*)SNESMEM(reg.S) = (byte)(reg.PC >> 16); (*(word*)&reg.S)--;
				*(word*)SNESMEM(reg.S-1) = (word)(reg.PC); (*(word*)&reg.S) -= 2;
				*(byte*)SNESMEM(reg.S) = (byte)(reg.P); (*(word*)&reg.S)--;
					// Push PC and P
				reg.P |= 4; // Interrupt flag
				reg.PC = x;
			}
		}
	}
	if (screen_scanline == scans_before_vbl()) {
		startvblank ();  // After the frame is done, do a VBL int
	}
	dohdmaline ();
	if (!debugmode && screen_scanline < scans_before_vbl() && frameskipwait < 0) {
		drawnextscanline ();
	}
}

#define MAXDEBUGHISTORY 20
#define MAXDEBUGLINELEN 40

char *debug_input(int &scancode)
{   // gets a debug command from user.
	// if scancode != 0, special key was pressed and returned
	static char line [MAXDEBUGHISTORY][MAXDEBUGLINELEN];
		// [0]=most recent line
	static int history = 0;
	int key, x, seekline = 0;

	scancode = 0;
	for (;;) {
		while (!kbhit()) {
			drawbox (0, 238, 255, 248, 0);
			setcolor (30); printf8 (0, 240, "Command:");
			setcolor (63); printf8 (50, 240, "%s%c", line[0], (tmsec & 128) ? '_' : ' ');
			copyscreen ();
		}
		key = getch ();
		if (key == 8 && strlen (line[0]) > 0)
			line[0][strlen(line[0])-1] = '\0';
		if (key == 13) {
			if (line [0][0] == '\0')
				return line[0];
			for (x = MAXDEBUGHISTORY-1; x > 0; x--) {
				strcpy (line[x], line[x-1]);
			}
			if (history < MAXDEBUGHISTORY)
				history++;
			line [0][0] = '\0';
			return line [1];
		}
		if (key >= 32 && key <= 127) {
			x = strlen (line[0]);
			if (x < MAXDEBUGLINELEN) {
				line[0][x] = toupper (key);
				line[0][x + 1] = '\0';
			}
		}
		if (key == 0) {
			key = getch ();
			switch (key) {
			case 'H': // up
				if (seekline < MAXDEBUGHISTORY - 1) {
					seekline++;
					strcpy (line[0], line[seekline]);
				}
				break;
			case 'P': // down
				if (seekline > 0) {
					seekline--;
					strcpy (line[0], line[seekline]);
				}
				break;
			}
			if ((key >= 59 && key <= 68) || key == 87 || key == 88) {
				scancode = key;
				return line[0];
			}
		}
		if (key == 27) {
			scancode = key;
			return line[0];
		}
	}
}

int grabnumber (char *line, int &i)
{
	int x;

	i = 0;
	while ((line[i] < '0' || line[i] > '9') && line[i] != '-' && line[i] != '$') {
		if (line[i] == '\0')
			return 0;
		i++;
	}
	if (line[i] == '$') {
		i++;
		x = 0;
		while ((line[i] >= '0' && line[i] <= '9') || (line[i] >= 'A' && line[i] <= 'F')) {
			x <<= 4;
			if (line[i] >= '0' && line[i] <= '9')
				x += line[i] - '0';
			if (line[i] >= 'A' && line[i] <= 'F')
				x += line[i] - 'A' + 10;
			i++;
		}
	} else {
		x = 0;
		while (line[i] >= '0' && line[i] <= '9') {
			x = x * 10 + (line[i] - '0');
			i++;
		}
	}
	return x;
}

void dodebugmode ()
{
	char *debugline, debuglinetext [55];
	boolean newinstr;
	dword opdata;
	int x, i, n, n2, esccount, key;
	FILE *fp;

	#define DEBUG_DOESCCHECK esccount = (esccount + 1) & 0x3F; if (esccount == 0 && kbhit() && getch () == 27) break

	(void) fp;
	newinstr = true;
	debugscreen = 1;
	applycustompalette ();
	while (!returntogui && debugmode) {
		if (newinstr) {
			newinstr = false;
			opdata = *(dword*)SNESMEM (reg.PC);
			unassemble (opdata, debuglinetext, reg.P);
			adddebugline (debuglinetext, true);
			applycustompalette ();
			if ((reg.A_dummy & 0xFFFF0000) || (reg.X_dummy & 0xFFFF0000) || (reg.Y_dummy & 0xFFFF0000)) {
				sprintf (debuglinetext, "Err! A$%0.8X X%0.8X Y%0.8X", reg.A_dummy, reg.X_dummy, reg.Y_dummy);
				adddebugline (debuglinetext, false, '@');
			}
			if ((reg.D_dummy & 0xFFFF0000) || (reg.S_dummy & 0xFFFF0000) || (reg.P_dummy & 0xFFFFFF00)) {
				sprintf (debuglinetext, "Err! D$%0.8X S%0.8X P%0.8X", reg.D_dummy, reg.S_dummy, reg.P_dummy);
				adddebugline (debuglinetext, false, '@');
			}
			if (reg.DBR_dummy1 || reg.DBR_dummy2 || reg.DBR_dummy3 || (reg.PC & 0xFF000000)) {
				sprintf (debuglinetext, "Err! D1$%0.2X D2$%0.2X D3$%0.2X PC$%0.8X", reg.DBR_dummy1, reg.DBR_dummy2, reg.DBR_dummy3, reg.PC);
				adddebugline (debuglinetext, false, '@');
			}
		}
		rendersnesscreen ();
		debugline = debug_input (key);
		if (key == 0) {
			if (debugline[0] == '\0') {
				debug_executecpu (0);
				newinstr = true;
			}
			for (x = 0; debugline[x] != '\0'; x++) {
				while (debugline[x] == ' ')
					x++;
				if (debugline [x] >= '0' && debugline[x] <= '9')
					debugscreen = debugline[x] - '0';
				switch (debugline [x]) {
				case 'T': // Trace
					n = grabnumber (debugline + x + 1, i);
					x += i;
					sprintf (debuglinetext, "[Tracing for %d instructions; ESC stops]", n);
					adddebugline (debuglinetext, false, '*');
					rendersnesscreen ();
					copyscreen ();
					while (n > 0 && !returntogui) {
						debug_executecpu(0);
						n--;
						DEBUG_DOESCCHECK;
					}
					newinstr = true;
					break;
				case 'W': // Wait for instruction
					while (debugline[x+1] == ' ')
						x++;
					for (n = 0; n < 256; n++) {
						if (stricmp (opcodelist[n].format, debugline + x+1) == 0) {
							break;
						}
					}
					if (n >= 256) {
						adddebugline ("[Format spec not found, waiting for string]", false, '*');
						do {
							debug_executecpu(0);
							DEBUG_DOESCCHECK;
							opdata = *(dword*)SNESMEM (reg.PC);
							unassemble (opdata, debuglinetext, reg.P);
							for (i = 0; *(debugline+x+1+i); i++)
								if (*(debugline+x+1+i) == '?' && debuglinetext[i] != '\0')
									debuglinetext[i] = '?';
						} while (strcmp (debugline+x+1, debuglinetext) != 0 && !returntogui);
						x += strlen (debugline + x+1);
						newinstr = true;
						break;
					}
					x += strlen (debugline + x+1);
					sprintf (debuglinetext, "[Tracing for opcode $%X; ESC stops]", n, opcodelist[n].format);
					adddebugline (debuglinetext, false, '*');
					rendersnesscreen ();
					copyscreen ();
					do {
						debug_executecpu(0);
						DEBUG_DOESCCHECK;
					} while (*(byte*)SNESMEM (reg.PC) != n && !returntogui);
					newinstr = true;
					break;
				case 'B': // break at address
					if (debugline[x+1] == 'N') { // break at next instruction
						n = reg.PC;
						x++;
						do {
							debug_executecpu(0);
							DEBUG_DOESCCHECK;
						} while (reg.PC != n+1 && reg.PC != n+2 && reg.PC != n+3 && reg.PC != n+4 && !returntogui);
						newinstr = true;
						break;
					}
					n = grabnumber (debugline + x + 1, i);
					x += i;
					if (debugline[x + 1] == '-') { // "Or Before..."
						x++;
						n2 = grabnumber (debugline + x + 1, i);
						x += i;
					} else
						n2 = 0;
					sprintf (debuglinetext, "[Tracing until PC=$%X; ESC stops]", n);
					adddebugline (debuglinetext, false, '*');
					rendersnesscreen ();
					copyscreen ();
					do {
						debug_executecpu(0);
						DEBUG_DOESCCHECK;
					} while (reg.PC < n-n2 || reg.PC > n && !returntogui);
					newinstr = true;
					break;
				case 'M':
					debugscreen = 3;
					while (debugline[x+1] == ' ')
						x++;
					if (debugline[x+1] == 'P' && debugline[x+2] == 'C') {
						memdumpoffset = reg.PC;
						x += 2;
					} else if (debugline[x+1] == 'S') {
						memdumpoffset = reg.S;
						x++;
					} else if (debugline[x+1] == 'D') {
						memdumpoffset = reg.D;
						x++;
					} else if (debugline[x+1] == 'X') {
						memdumpoffset = (reg.DBR << 16) + reg.X;
						x++;
					} else if (debugline[x+1] == 'Y') {
						memdumpoffset = (reg.DBR << 16) + reg.Y;
						x++;
					} else {
						memdumpoffset = grabnumber (debugline + x + 1, i);
						x += i;
					}
					break;
				case 'C': // break at memory change/trap
					if (debugline[x+1] == 'T') {
						x++;
						n = grabnumber (debugline + x + 1, i);
						x += i;
						sprintf (debuglinetext, "[Tracing until $%X($%X) is trapped; ESC stops]", n, *SNESMEM(n));
						adddebugline (debuglinetext, false, '*');
						rendersnesscreen ();
						copyscreen ();
						debug_trapaddr = -1;
						while (debug_trapaddr != n && !returntogui) {
							opdata = *(dword*)SNESMEM (reg.PC);
							debug_executecpu(0);
							DEBUG_DOESCCHECK;
						}
						unassemble (opdata, debuglinetext, reg.P);
						adddebugline (debuglinetext, false, '*');
						sprintf (debuglinetext, "Value now: $%X  Read trap? %d", *SNESMEM(n), debug_readtrap);
						adddebugline (debuglinetext, false, '*');
						newinstr = true;
						break;
					} else if (debugline[x+1] == 'V') {
						x++;
						if (debugline[x+1] == 'P') {
							x++;
							n = grabnumber (debugline + x + 1, i);
							x += i;
							sprintf (debuglinetext, "[Tracing until VRAM pointer is set to $%X; ESC stops]", n);
							adddebugline (debuglinetext, false, '*');
							rendersnesscreen ();
							copyscreen ();
							while (state.vrampointer != n && !returntogui) {
								debug_executecpu(0);
								DEBUG_DOESCCHECK;
							}
							newinstr = true;
							break;
						} else {
							n = grabnumber (debugline + x + 1, i);
							x += i;
							sprintf (debuglinetext, "[Tracing until VRAM at $%X($%X) changes; ESC stops]", n, vram[n]);
							adddebugline (debuglinetext, false, '*');
							rendersnesscreen ();
							copyscreen ();
							i = vram[n];
							while (vram[n] == i && !returntogui) {
								debug_executecpu(0);
								DEBUG_DOESCCHECK;
							}
							newinstr = true;
							break;
						}
					} else if (debugline[x+1] == 'C' && debugline[x+2] == 'P') {
						x += 2;
						n = grabnumber (debugline + x + 1, i);
						x += i;
						sprintf (debuglinetext, "[Tracing until CGRAM pointer is set to $%X; ESC stops]", n);
						adddebugline (debuglinetext, false, '*');
						rendersnesscreen ();
						copyscreen ();
						while (state.colorpointer != n && !returntogui) {
							debug_executecpu(0);
							DEBUG_DOESCCHECK;
						}
						newinstr = true;
						break;
					}
					n = grabnumber (debugline + x + 1, i);
					x += i;
					i = *SNESMEM(n);
					sprintf (debuglinetext, "[Tracing until $%X($%X) changes; ESC stops]", n, i);
					adddebugline (debuglinetext, false, '*');
					rendersnesscreen ();
					copyscreen ();
					while (*SNESMEM(n) == i) {
						opdata = *(dword*)SNESMEM (reg.PC);
						debug_executecpu(0);
						DEBUG_DOESCCHECK;
					}
					unassemble (opdata, debuglinetext, reg.P);
					adddebugline (debuglinetext, false, '*');
					newinstr = true;
					break;
				case 'V':
					if (debugline[x+1] == 'B') {
						debugscreen = 8;
						x++;
						vramblitoffset = grabnumber (debugline + x + 1, i);
						x += i;
						vramblitcdepth = grabnumber (debugline + x + 1, i);
						x += i;
						break;
					} else if (debugline[x+1] == 'F') {
						x++;
						if (debugline[x+1] == 'C') { // Copy (swap!)
							x++;
							for (n = 0; n < 0x10000; n++) {
								n2 = vram[n];
								vram[n] = debug_vram[n];
								debug_vram[n] = n2;
							}
							memset (cstatus, 0, sizeof (cstatus));
							break;
						} else if (debugline[x+1] == 'S') { // Save
							if ((fp = fopen (vramcmpfn, "wb")) == NULL) {
								adddebugline ("Can't save VRAM compare file", false, '*');
								break;
							}
							x++;
							fwrite (debug_vram, 1, 0x10000, fp);
							fclose (fp);
							break;
						}
						if (debugline[x+1] != 'R') { // R=Reload
							strcpy (vramcmpfn, debugline + x + 1);
							x += strlen (debugline + x + 1);
							if (vramcmpfn[0] == '\0')
								break;
						} else
							x++;
						if ((fp = fopen (vramcmpfn, "rb")) == NULL) {
							adddebugline ("Can't load VRAM compare file", false, '*');
							break;
						}
						fread (debug_vram, 1, 0x10000, fp);
						fclose (fp);
						break;
					}
					debugscreen = 4;
					vramdumpoffset = grabnumber (debugline + x + 1, i);
					x += i;
					break;
				case 'O':
					debugscreen = 6;
					oamdumpoffset = grabnumber (debugline + x + 1, i);
					x += i;
					break;
				case 'N':
					if (debugline[x+1] == 'N') {
						i = *REG4200;
						adddebugline ("Calling NMI Interrupt Vector", false, i & 0x80 ? '+' : '-');
						*REG4200 |= 0x80;
						startvblank ();
						*REG4200 = i;
						newinstr = true;
						break;
					} else if (debugline[x+1] == 'L') {
						adddebugline ("[Tracing until VBlank begins; ESC stops]", false, '*');
						rendersnesscreen ();
						copyscreen ();
						if (screen_scanline >= scans_before_vbl()) {
							do {
								debug_executecpu(0);
								DEBUG_DOESCCHECK;
							} while (screen_scanline >= scans_before_vbl() && !returntogui);
						}
						while (screen_scanline < scans_before_vbl() && !returntogui) {
							debug_executecpu(0);
							DEBUG_DOESCCHECK;
						}
						newinstr = true;
						break;
					}
					adddebugline ("[Tracing until NMI call; ESC stops]", false, '*');
					rendersnesscreen ();
					copyscreen ();
					debug_nmistart = false;
					while (!debug_nmistart && !returntogui) {
						debug_executecpu(0);
						DEBUG_DOESCCHECK;
					}
					newinstr = true;
					break;
				case 'R':
					x++;
					while (debugline[x] == ' ')
						x++;
					switch (debugline[x]) {
					case 'P':
						if (debugline[x+1] == 'C') {
							x++;
							n = grabnumber (debugline + x + 1, i);
							x += i; reg.PC = n;
						} else {
							n = grabnumber (debugline + x + 1, i);
							x += i; reg.P = n;
						}
						break;
					case 'D':
						if (debugline[x+1] == 'B') {
							x++;
							n = grabnumber (debugline + x + 1, i);
							x += i; reg.DBR = n;
						} else {
							n = grabnumber (debugline + x + 1, i);
							x += i; reg.D = n;
						}
						break;
					case 'A':
						n = grabnumber (debugline + x + 1, i);
						x += i; reg.A = n;
						break;
					case 'X':
						n = grabnumber (debugline + x + 1, i);
						x += i; reg.X = n;
						break;
					case 'Y':
						n = grabnumber (debugline + x + 1, i);
						x += i; reg.Y = n;
						break;
					case 'S':
						if (debugline[x+1] == 'C') {
							x++;
							n = grabnumber (debugline + x + 1, i);
							x += i; scan_cycles = n;
						} else if (debugline[x+1] == 'L') {
							x++;
							n = grabnumber (debugline + x + 1, i);
							x += i; screen_scanline = n;
							frame_cycles = (261-screen_scanline) * cycles_per_scan;
						} else {
							n = grabnumber (debugline + x + 1, i);
							x += i; reg.S = n;
						}
						break;
					case 'I':
						if (debugline[x+1] == 'C') {
							x++;
							n = grabnumber (debugline + x + 1, i);
							x += i; debug_instr = n;
						}
						break;
					}
					adddebugline ("Register set", true, '*');
					break;
				case 'S':
					n2 = grabnumber (debugline + x + 1, i);
					x += i;
					if (debugline [x + 1] == '\0') {
						trapregread (SNESMEM (n2), false);
						break;
					}
					n = grabnumber (debugline + x + 1, i);
					x += i;
					if (i > 4 || n >= 256) {
						*(word*)SNESMEM(n2) = n;
						trapregread (SNESMEM (n2), true);
						trapregwrite (SNESMEM (n2), true);
					} else {
						*(byte*)SNESMEM(n2) = n;
						trapregread (SNESMEM (n2), false);
						trapregwrite (SNESMEM (n2), false);
					}
					break;
				case '?':
					adddebugline ("Press F9 to Run (exit debug mode)", false, '-');
					adddebugline ("Tx: Trace for x instructions ($=Hex)", false, '-');
					adddebugline ("Bx: Break at 24-bit address [BN=break@next]", false, '-');
					adddebugline ("Wx: Wait for op eg. W LDX $4210 (slow!)", false, '-');
					adddebugline ("Cx: Break when memory is changed at address", false, '-');
					adddebugline ("C[T/V/VP/CP]x: Check address x for [MMIO Trap/", false, '-');
					adddebugline ("   VRAM modify/VRAM Pointer set]", false, '-');
					adddebugline ("Mx: Dump SNES memory at address, PC, S, etc.", false, '-');
					adddebugline ("Ox: Dump OAM memory at address", false, '-');
					adddebugline ("Vx: Dump VRAM memory at address", false, '-');
					adddebugline ("VBo d: Tile Blit o=address d=color depth 0/1/3", false, '-');
					adddebugline ("N[N|L]: Wait for NMI int. call [Call now/VBL line]", false, '-');
					adddebugline ("Rrv: Set register/SC/SL/IC eg R A $1234", false, '-');
					adddebugline ("Sxv: Set memory eg S $1234 $56 (trapped)", false, '-');
					adddebugline ("0 to 7: Choose a debug screen (1=Normal)", false, '-');
					adddebugline ("SPx y: Debug screen 0 sprite range x to y", false, '-');
					adddebugline ("PG or PD: Choose palette (game or debug)", false, '-');
					break;
				case 'P':
					x++;
					if (debugline[x] == 'G') {
						for (i = 0; i < 256; i++) {
							setsnespalette (i, ((word*)cgram) [i], (*REG2100 & 0xF));
						}
					}
					if (debugline[x] == 'D')
						applycustompalette ();
					break;
				case '\\':
					debug_instr = grabnumber (debugline + x + 1, i);
					x += i;
					break;
				}
			}
		} else if (key == 27) {
			returntogui = true;
		} else if (key != 0) {
			if (key >= 59 && key <= 62)
				curgs->bgs ^= 1 << (key - 59);
			switch (key) {
			case 63: // F5
				instantsave (instantsaveslot); break;
			case 64: // F6
				memset (cstatus, 0, sizeof (cstatus));
				break;
			case 65: // F7
				instantload (instantsaveslot);
				newinstr = true;
				break;
			case 66: // F8
				curgs->bgs ^= 0x10; break;
			case 67: // F9
				debugmode = false;
				break;
			}
		}
	}
	adddebugline ("[Debugger exitted]", false, '*');
	rendersnesscreen ();
	applygamepalette ();
	debugmode = false;
}

void togglecheat (int n)
{
	int x;
	struct cheatpatchset *cheat = curgs->cheat[n];
	byte *data;

	if (cheat->enabled) {
		cheat->enabled = false;
		for (x = 0; x < cheat->addrs; x++) {
			data = SNESMEM ((cheat->patch[x].bank << 16) + cheat->patch[x].addr);
			if (cheat->patch[x].wordsized) {
				*(word*)data = cheat->patch[x].prevval;
			} else {
				*data = (byte) cheat->patch[x].prevval;
			}
		}
	} else {
		cheat->enabled = true;
		for (x = 0; x < cheat->addrs; x++) {
			data = SNESMEM ((cheat->patch[x].bank << 16) + cheat->patch[x].addr);
			if (cheat->patch[x].wordsized) {
				cheat->patch[x].prevval = *(word*)data;
				*(word*)data = cheat->patch[x].newval;
			} else {
				cheat->patch[x].prevval = *data;
				*data = (byte) cheat->patch[x].newval;
			}
		}
	}
}

size_t vfwrite(void *buf, size_t elsize, size_t nelem, FILE *fp)
{
	dword abf = (dword)fp-1;
	if (abf < 10) {
		memcpy (autobacktrackmem[abf] + autobackpos[abf], buf, elsize*nelem);
		autobackpos[abf] += elsize*nelem;
		assert (autobackpos[abf] < AUTOBACKFILESIZE);
		return elsize*nelem;
	}
	return fwrite (buf, elsize, nelem, fp);
}

size_t vfread(void *buf, size_t elsize, size_t nelem, FILE *fp)
{
	dword abf = (dword)fp-1;
	if (abf < 10) {
		memcpy (buf, autobacktrackmem[abf] + autobackpos[abf], elsize*nelem);
		autobackpos[abf] += elsize*nelem;
		assert (autobackpos[abf] < AUTOBACKFILESIZE);
		return elsize*nelem;
	}
	return fread (buf, elsize, nelem, fp);
}

int vfeof (FILE *fp)
{
	dword abf = (dword)fp-1;
	if (abf < 10) {
		if (autobackpos[abf] >= autobacksize[abf])
			return true;
		else
			return false;
	}
	return feof (fp);
}

FILE *vfopen (const char *filename, const char *mode)
{
	if (filename[0] >= '0' && filename[0] <= '9' && filename[1] == '\0') {
		autobackpos [filename[0] - '0'] = 0;
		if (autobacktrackmem[filename[0] - '0'] == NULL) return NULL;
		return (FILE*)(filename[0] - '0' + 1);
	}
	return fopen (filename, mode);
}

void vfclose (FILE *fp)
{
	if ((dword)fp > 10) {
		fclose (fp);
	} else {
		autobacksize[(int)fp - 1] = autobackpos[(int)fp - 1];
		total_frames = 1; // To avoid confusing the auto-backtracker.
	}
}

inline int autobackfilesize()
{
	return AUTOBACKFILESIZE + (curgs->emulatespc ? AUTOBACKSPCSIZE : 0);
	// Approximate size of quick save file
}

void alloc_autobacktrack(int num)
{
	int x;
	for (x = 0; x < 9; x++) {
		if (x < num) { // allocate
			if (!autobacktrackmem[x]) {
				autobacktrackmem[x] = (byte*) malloc (autobackfilesize());
				if (autobacktrackmem[x] == NULL) {
					addmessage ("ERROR!  Out of memory for auto-backtrack file(s)!");
					return;
				}
				autobacktrackmem[x][0] = 0; // Mark unused
			}
		} else if (autobacktrackmem[x]) { // deallocate
			free(autobacktrackmem[x]);
			autobacktrackmem[x] = NULL;
		}
	}
	if (num != 0)
		addmessage ("%dKB reserved for auto-backtrack.", autobackfilesize() * num / 1024);
}

void dealloc_autobacktrack()
{
	alloc_autobacktrack (0);
	autoback_restorecount = 0;
	autoback_countdown = 0;
}

void restorebacktrack ()
{
	int x;

	if (autoback_restorecount < 60) {
		// Backtrack was restored before within the last second, so delete the current file and load the previous
		for (x = 0; x < 9; x++) {
			if (autobacktrackmem[x] && autobacktrackmem[x + 1]) {
				memcpy (autobacktrackmem[x], autobacktrackmem[x + 1], autobackfilesize());
			} else if (autobacktrackmem[x]) {
				autobacktrackmem[x][0] = 0;
			}
		}
	}
	if (autobacktrackmem[0] && autobacktrackmem[0][0] != 0) {
		instantload (0, true);
		autoback_countdown = backtracktime * 60;
		autoback_restorecount = 0;
	} else {
		return;
	}
	if (autobacktrackmem[0] && autobacktrackmem[0][0] == 0)
		instantsave (0, true); // Make sure there's something for the player to backtrack to.
}

void savebacktrack ()
{
	int x;
	for (x = 8; x > 0; x--) {
		if (autobacktrackmem[x] && autobacktrackmem[x - 1]) {
			memcpy (autobacktrackmem[x], autobacktrackmem[x - 1], autobackfilesize());
		}
	}
	if (autobacktrackmem[0])
		instantsave (0, true);
}

inline int getoptimalnframes (int msec)
{
	return msec / (ispal() ? 2000 / curgs->regpercent : 1700 / curgs->regpercent);
}

void runsnes ()
{
	boolean checkjoy = false, wasrendered = false;
	int vfps, rfps;
	int n, x, y, key;
	byte *data;
	struct {
		int msec;
		int nframes;
		int optimalnframes;
		int oldtmsec;
	} autoreg;

	frameskipwait = 0;
	romrunning = true;
		// This is set to false right after a ROM is loaded, and right after the system
		// gets reset.  The user is only allowed to change Startup settings while the
		// ROM is not running.
	setvideomode (gameres, true);
		// Set the video mode to the one selected by the user (gameres=game resolution);
		// The virtual resolution is set to 256x256x256.  The virtual screen memory
		// is stored in the global variable vs[].
	memcpy (vs, snesscreen, 0x10000);
	copyscreen();

	returntogui = false;
	messageoutline = false;
	for (x = 0; x < FTHISTORY; x++)
		frametime [x] = 1667/curgs->regpercent;
	for (x = 0; x < sizeof (struct snesjoypad); x++) {
		if (((byte *)&joykey [pcontrol[0]]) [x] >= 100) {
			checkjoy = true; break;
		}
		if (((byte *)&joykey [pcontrol[1]]) [x] >= 100) {
			checkjoy = true; break;
		}
	}
	autoreg.msec = 1000; // Reset auto-regulation values, assume optimal speed already
	autoreg.nframes = getoptimalnframes (autoreg.msec);
	applygamepalette ();
	while (!returntogui) {
		if (debugmode) {
			dodebugmode ();
			continue;
		}
		if (curgs->autoreg) {
			autoreg.oldtmsec = tmsec; // Start timing this frame...
		}
		framestarttime = tmsec;
		debugscreen = 0;
		if (checkjoy) {
			checkjoystick ();
		}
		while (kbhit ()) {
			key = getch ();
			debug0 ("Read key: %d (%c)", key, key);
			if (key == 27)
				returntogui = true;
			else if (key == '+' || key == '=') {
				if (curgs->autoreg) {
					curgs->regpercent += 10;
					if (curgs->regpercent > 200)
						curgs->regpercent = 200;
					addmessage ("Speed: %d%%", curgs->regpercent);
					autoreg.msec = 1000; // Reset values, assume optimal speed already
					autoreg.nframes = getoptimalnframes (autoreg.msec);
				} else {
					if (curgs->frameskip < 9) {
						curgs->frameskip++;
					}
				}
			} else if (key == '-' || key == '_') {
				if (curgs->autoreg) {
					curgs->regpercent -= 10;
					if (curgs->regpercent < 10)
						curgs->regpercent = 10;
					addmessage ("Speed: %d%%", curgs->regpercent);
					autoreg.msec = 1000; // Reset auto-regulation values, assume optimal speed already
					autoreg.nframes = getoptimalnframes (autoreg.msec);
				} else {
					if (curgs->frameskip > 0) {
						curgs->frameskip--;
					}
				}
			}
			if (key >= '1' && key <= '5')
				curgs->bgs ^= 1 << (key - '1');
			if (key == 0) {
				key = getch();
				debug0 ("Extended key %d %c", key, key);
				if (key >= 59 && key <= 61)
					curgs->bgs ^= 1 << (key - 59);
				switch (key) {
				case 62: // F4
					showfps = !showfps;
					break;
				case 63: // F5
					if (!disallowinst)
						instantsave (instantsaveslot); break;
				case 64: // F6
					curgs->bgs = 0x1F; // all BGs and sprites
					break;
				case 65: // F7
					if (!disallowinst)
						instantload (instantsaveslot); break;
				case 66: // F8
					curgs->bgs ^= 0x10; break;
				case 67: // F9
					debugmode = true; break;
				case 68: // F10
					//soundchannels = 0xFF; break;
					restorebacktrack ();
				case 87: // F11
					soundchannels = 0; break;
				}
			}
		}
		for (x = 0; x < 8; x++)
			if (curgs->emulatespc && keydown [x + 2] && keydown [KEYCONTROL]) {
				soundchannels ^= 1 << x;
			}
		for (x = 1; x <= 10; x++)
			if (keydown [x + 1] && keydown [KEYALT]) {
				instantsaveslot = x % 10;
				addmessage ("Instant save slot %d selected.", instantsaveslot);
			}

		// frame cycles initted to 0, so loop is skipped the first time.
		if (!curgs->autoreg) {
			frameskipwait--;
		}
		if (frameskipwait < 0) {
			clearscreen (0);
			wasrendered = true;
			// frame is rendered only when frameskipwait < 0
		}
		// Apply cheats every frame
		for (n = 0; n < curgs->patches; n++) {
			if (curgs->cheat[n]->enabled) {
				for (x = 0; x < curgs->cheat[n]->addrs; x++) {
					data = SNESMEM ((curgs->cheat[n]->patch[x].bank << 16) \
						       + curgs->cheat[n]->patch[x].addr);
					if (curgs->cheat[n]->patch[x].wordsized) {
						*(word*)data = curgs->cheat[n]->patch[x].newval;
					} else {
						*data = (byte) curgs->cheat[n]->patch[x].newval;
					}
				}
			}
		}
		while (frame_cycles > 0) {
			// Begginning of screen scanline
			startscanline ();
			while (scan_cycles >= 0) {
				//x = scan_cycles;
				//scan_cycles = 0;
				executecpu ();
				//scan_cycles = x + scan_cycles;
			}
			frame_cycles -= cycles_per_scan;
			scan_cycles += cycles_per_scan;
			screen_scanline++;
			//if (screen_scanline < 10 || screen_scanline > 220)
			//	debug0 ("scanline %d new scan_cycles %d frame_cycles %d PC$%X", screen_scanline, scan_cycles, frame_cycles, reg.PC);
		}
		startcpuframe ();
		total_frames++;
		autoback_countdown--;
		autoback_restorecount++;
		debug0 ("Frame %d: debug_instr=%d", total_frames, debug_instr);

		if (autobacktrack && autoback_countdown <= 0) {
			savebacktrack();
			autoback_countdown = backtracktime * 60;
		}

		if (frameskipwait < 0) {
			getfps (vfps, rfps);
			if (!returntogui) {
				if ((snesrenderedframe & 1) == 0)
					messagecolor = findclosestmatch (0xFFFFE0);
				showmessages ();
				if (showfps) {
					setcolor (messagecolor);
					printf8 (215, 234, "%d.%d/%d.%d", rfps / 10, rfps % 10, vfps / 10, vfps % 10);
					if (curfastrom) printf8 (215, 245, "FastROM");
					// Debug vars
					//printf8 (200, 228, "%0.4X/%0.4X/%0.4X", *(word*)SNESMEM(0x7E000D), *(word*)SNESMEM(0x7E0096), *(word*)SNESMEM(0x7E00D3));
				}
			}
			for (x = 0; x < 4; x++) {
				if (!(curgs->bgs & (1 << x))) {
					setcolor (messagecolor);
					printf8 (0, 200 + (x << 3), "%d", x + 1);
				}
			}
			if (!(curgs->bgs & 0x10)) {
				setcolor (messagecolor);
				printf8 (0, 232, "S");
			}
			if (showfps) {
				// Debug thing: show sprite # when mouse is over it.
				for (n = 0; n < 128; n++) {
					int size = spr8x8size[n] << 3;
					y = oam [(n << 2) + 1] + 8;
					if (mousey < y + size && mousey >= y) {
						x = oam [(n << 2)] + ((oam [512+(n>>2)] & (1<<((n&3)<<1))) ? -256 : 0);
						if (mousex >= x && mousex < x + size) {
							setcolor (messagecolor);
							printf8 (n, 0, "%d", n);
							drawdottedlinebox (x, y, x + size - 1, y + size - 1, messagecolor);
						}
					}
				}
				checkmouse ();
				drawmouse ();
			}
			copyscreen ();

			if (!curgs->autoreg) {
				frameskipwait = curgs->frameskip;
				while (frameskipwait < 0) {
					// If curframeskip was < 0, induce a deliberate delay
					frameskipwait++;
					rendersnesscreen ();
				}
			}
		} else {
			wasrendered = false;
		}
		if (curgs->autoreg) {
			static int skipped = 0;
			
			autoreg.nframes++;
			autoreg.msec += tmsec - autoreg.oldtmsec;
			x = getoptimalnframes (autoreg.msec);
			//debug0 ("frame took %dms total %dms, %d frames; optimal frames %d", tmsec - autoreg.oldtmsec, autoreg.msec, autoreg.nframes, x);
			if (autoreg.nframes < x) {
				frameskipwait = 1;
				skipped++;
				if (skipped >= 10) {
					frameskipwait = -1;
					skipped = 0; // Force a frame render
				}
			} else {
				frameskipwait = -1;
				skipped = 0;
				if (x + 5 <= autoreg.nframes) { // Game running too fast
					for (n = 0; n < (autoreg.nframes - x) / 5; n++) {
						rendersnesscreen (); // Just to waste CPU time
					}
				}
			}
			
		}
		memmove (frametime, frametime + 1, FTHISTORY - 1);
		memmove (framewasrendered, framewasrendered + 1, FTHISTORY - 1);
		framewasrendered [FTHISTORY - 1] = wasrendered;
		frametime [FTHISTORY - 1] = tmsec - framestarttime;
	}

	memcpy (snesscreen, vs, 0x10000);
		// Save the snes screen so that you can save a screen snapshot in the GUI,
		// and so that you can use the game and dim game backgrounds.
	savegamepalette ();
}
